#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_active_eyes(var, tint, mms_id):
    r"""This function returns the FEEPS active eyes,
    based on date/mms_id/species/rate.

    Parameters
    ----------
    var : dict
        Hash table containing parameters.
    tint : list of str
        Time range.
    mms_id : int or str
        mms_id e.g., '4' for MMS4.

    Returns
    -------
    sensors : dict
        Hash table containing 2 keys :
            * out["top"] : maps to the active top eyes.
            * out["bottom"] : maps to the active bottom eyes.

    Notes
    -----
    1) Burst mode should include all sensors (TOP and BOTTOM) :
        * electrons : [1, 2, 3, 4, 5, 9, 10, 11, 12].
        * ions : [6, 7, 8].

    2) SITL should return (TOP only) :
        * electrons : set_intersection([5, 11, 12], active_eyes).
        * ions : None.

    3) From Drew Turner, 9/7/2017, srvy mode :
        * before 16 August 2017 :
            * electrons : [3, 4, 5, 11, 12].
            * ion : [6, 7, 8].

        * after 16 August 2017 :
            * MMS1 :
                * Top Eyes : 3, 5, 6, 7, 8, 9, 10, 12
                * Bot Eyes : 2, 4, 5, 6, 7, 8, 9, 10
            * MMS2 :
                * Top Eyes : 1, 2, 3, 5, 6, 8, 10, 11
                * Bot Eyes : 1, 4, 5, 6, 7, 8, 9, 11
            * MMS3
                * Top Eyes : 3, 5, 6, 7, 8, 9, 10, 12
                * Bot Eyes : 1, 2, 3, 6, 7, 8, 9, 10
            * MMS4 :
                * Top Eyes : 3, 4, 5, 6, 8, 9, 10, 11
                * Bot Eyes : 3, 5, 6, 7, 8, 9, 10, 12

    """

    if isinstance(mms_id, str):
        mms_id = int(mms_id)

    sensors = {}

    if var["tmmode"].lower() == "brst" and var["dtype"].lower() == "electron":
        sensors["top"] = [1, 2, 3, 4, 5, 9, 10, 11, 12]
        sensors["bottom"] = [1, 2, 3, 4, 5, 9, 10, 11, 12]
        return sensors

    if var["tmmode"].lower() == "brst" and var["dtype"].lower() == "ion":
        sensors["top"] = [6, 7, 8]
        sensors["bottom"] = [6, 7, 8]
        return sensors

    # old eyes, srvy mode, prior to 16 August 2017
    if var["dtype"].lower() == "electron":
        sensors["top"] = [3, 4, 5, 11, 12]
        sensors["bottom"] = [3, 4, 5, 11, 12]
    else:
        sensors["top"] = [6, 7, 8]
        sensors["bottom"] = [6, 7, 8]

    if isinstance(tint[0], str):
        start_day = np.array(tint[0]).astype("<M8[D]")
    else:
        start_day = tint[0].astype("<M8[D]")

    # srvy mode, after 16 August 2017
    if start_day >= np.datetime64("2017-08-16") and var["tmmode"].lower() == "srvy":
        active_table = {
            "1-electron": {},
            "1-ion": {},
            "2-electron": {},
            "2-ion": {},
            "3-electron": {},
            "3-ion": {},
            "4-electron": {},
            "4-ion": {},
        }

        active_table["1-electron"]["top"] = [3, 5, 9, 10, 12]
        active_table["1-electron"]["bottom"] = [2, 4, 5, 9, 10]

        active_table["1-ion"]["top"] = [6, 7, 8]
        active_table["1-ion"]["bottom"] = [6, 7, 8]

        active_table["2-electron"]["top"] = [1, 2, 3, 5, 10, 11]
        active_table["2-electron"]["bottom"] = [1, 4, 5, 9, 11]

        active_table["2-ion"]["top"] = [6, 8]
        active_table["2-ion"]["bottom"] = [6, 7, 8]

        active_table["3-electron"]["top"] = [3, 5, 9, 10, 12]
        active_table["3-electron"]["bottom"] = [1, 2, 3, 9, 10]

        active_table["3-ion"]["top"] = [6, 7, 8]
        active_table["3-ion"]["bottom"] = [6, 7, 8]

        active_table["4-electron"]["top"] = [3, 4, 5, 9, 10, 11]
        active_table["4-electron"]["bottom"] = [3, 5, 9, 10, 12]

        # active_table["4-electron"]["top"] = [3, 4, 5, 10, 11]
        # active_table["4-electron"]["bottom"] = [3, 5, 9, 10, 12]

        active_table["4-ion"]["top"] = [6, 8]
        active_table["4-ion"]["bottom"] = [6, 7, 8]

        sensors = active_table[f"{mms_id:d}-{var['dtype'].lower()}"]

        if var["lev"].lower() == "sitl":
            sensors["top"] = list(set(sensors["top"]) & {5, 11, 12})
            sensors["bottom"] = []
            return {
                "top": list(set(sensors["top"]) & {5, 11, 12}),
                "bottom": [],
            }

    if var["lev"].lower() == "sitl":
        sensors["top"] = [5, 11, 12]
        sensors["bottom"] = []

    return sensors



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from datetime import datetime
from typing import Union

# 3rd party imports
import numpy as np
import pandas as pd
from numpy.typing import NDArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def datetime2iso8601(
    time: Union[list[datetime], NDArray[datetime], datetime],
) -> Union[list[str], str]:
    r"""Transforms datetime to TT2000 string format.

    Parameters
    ----------
    time : datetime
        Time to convert to tt2000 string.

    Returns
    -------
    tt2000 : str
        Time in TT20000 iso_8601 format.

    """

    if isinstance(time, (np.ndarray, list)):
        time_iso8601 = []

        for t in time:
            time_iso8601.append(datetime2iso8601(t))

    elif isinstance(time, datetime):
        time_datetime = pd.Timestamp(time)

        # Convert to string
        datetime_str = time_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")

        time_iso8601 = f"{datetime_str}{time_datetime.nanosecond:03d}"

    else:
        raise TypeError("time must be array_like or datetime")

    return time_iso8601



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from pycdfpp import load

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.iso86012datetime64 import iso86012datetime64
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_skymap import ts_skymap
from .get_ts import _get_epochs
from .get_variable import _pycdfpp_attributes_to_dict

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

# Keys of the global attributes to keep from CDF informations
Globkeys = [
    "CDF",
    "Version",
    "Encoding",
    "Checksum",
    "Compressed",
    "LeapSecondUpdated",
]


def get_dist(file_path, cdf_name, tint: list = None):
    r"""Read field named cdf_name in file and convert to velocity distribution
    function.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.
    tint : list of str, Optional
        Time interval.

    Returns
    -------
    out : xarray.Dataset
        Time series of the velocity distribution function if the target
        specie in the selected time interval.

    """

    tmmode = cdf_name.split("_")[-1]

    if "_dis_" in cdf_name:
        specie = "ions"
    elif "_des_" in cdf_name:
        specie = "electrons"
    else:
        raise AttributeError(
            "Couldn't get the particle species from file name!!",
        )

    # Check time interval type
    # Check time interval
    if tint is None:
        tint = ["1995-10-06T18:50:00.000000000", "2200-10-06T18:50:00.000000000"]
    elif isinstance(tint, (np.ndarray, list)):
        if isinstance(tint[0], np.datetime64):
            tint = datetime642iso8601(np.array(tint))
        elif isinstance(tint[0], str):
            tint = iso86012datetime64(
                np.array(tint),
            )  # to make sure it is ISO8601 ok!!
            tint = datetime642iso8601(np.array(tint))
        else:
            raise TypeError("Values must be in datetime64, or str!!")
    else:
        raise TypeError("tint must be array_like!!")

    # Load CDF file
    file = load(file_path)

    # with CDF(file_path) as file:
    # Get the relevant CDF file information (zVariables)
    z_vars = [z_var[0] for z_var in file.items()]

    # Get the global attributes
    glob_attrs = _pycdfpp_attributes_to_dict(file.attributes)
    glob_attrs = {**glob_attrs, **{"tmmode": tmmode, "species": specie}}

    # Get VDF zVariable attributes
    dist_attrs = _pycdfpp_attributes_to_dict(file[cdf_name].attributes)

    # Get CDF keys to Epoch, energy, azimuthal and elevation angle
    # zVariables
    depends_keys = [dist_attrs[f"DEPEND_{i:d}"] for i in range(4)]

    # Get coordinates attributes
    coords_attrs = {}

    for n, k in zip(["time", "phi", "theta", "energy"], depends_keys):
        coords_attrs[n] = _pycdfpp_attributes_to_dict(file[k].attributes)

    times = _get_epochs(file, cdf_name)

    # If something time is None means that there is nothing interesting
    # in this file so leave!!
    if times["data"] is not None:
        times = times["data"]
    else:
        return None

    dist = np.transpose(file[cdf_name].values, [0, 3, 1, 2])
    phi, theta, energy = [np.squeeze(file[k].values) for k in depends_keys[1:]]

    if tmmode == "brst":
        en0_name = "_".join(
            [
                cdf_name.split("_")[0],
                cdf_name.split("_")[1],
                "energy0",
                cdf_name.split("_")[-1],
            ],
        )
        en1_name = "_".join(
            [
                cdf_name.split("_")[0],
                cdf_name.split("_")[1],
                "energy1",
                cdf_name.split("_")[-1],
            ],
        )

        e_step_table_name = "_".join(
            [
                cdf_name.split("_")[0],
                cdf_name.split("_")[1],
                "steptable_parity",
                cdf_name.split("_")[-1],
            ],
        )

        step_table = file[e_step_table_name].values

        if en0_name not in z_vars:
            if energy.ndim == 1:
                energy0 = energy
                energy1 = energy
            elif energy.shape[0] == 1:
                energy0 = energy[0, :]
                energy1 = energy[0, :]
            else:
                idx_energy0 = np.where(step_table == 0)[0]
                idx_energy1 = np.where(step_table == 1)[0]
                energy0 = energy[idx_energy0[0], :]
                energy1 = energy[idx_energy1[0], :]
        else:
            energy0 = file[en0_name].values
            energy1 = file[en1_name].values

        # Overwrite energy to make sure that energy0 and energy1
        # are used instead
        energy = np.tile(energy0, (len(step_table), 1))
        energy[step_table == 1] = np.tile(energy1, (int(np.sum(step_table)), 1))

    elif tmmode == "fast":
        phi = np.tile(phi, (len(times), 1))

        if energy.ndim == 1:
            energy0 = energy
            energy1 = energy
        elif energy.shape[0] == 1:
            energy0 = energy[0, :]
            energy1 = energy[0, :]
        else:
            energy0 = energy[1, :]
            energy1 = energy[0, :]

        step_table = np.zeros(len(times))

    else:
        raise ValueError("Invalid sampling mode!!")

    d_en_name = "_".join(
        [
            cdf_name.split("_")[0],
            cdf_name.split("_")[1],
            "energy_delta",
            cdf_name.split("_")[-1],
        ],
    )

    if d_en_name in z_vars:
        glob_attrs["delta_energy_plus"] = file[d_en_name].values
        glob_attrs["delta_energy_minus"] = file[d_en_name].values
    else:
        glob_attrs["delta_energy_plus"] = None
        glob_attrs["delta_energy_minus"] = None

    out = ts_skymap(
        times,
        dist,
        energy,
        phi,
        theta,
        energy0=energy0,
        energy1=energy1,
        esteptable=step_table,
        attrs=dist_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    out = time_clip(out, tint)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

# Local imports
from .resample import resample
from .ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pres_anis(p_fac, b_xyz):
    r"""Compute pressure anisotropy factor:

    .. math::

        \mu_0 \frac{P_\parallel - P_\perp}{|\mathbf{B}|^2}

    Parameters
    ----------
    p_fac : xarray.DataArray
        Time series of the pressure tensor in field aligne coordinates.
    b_xyz : xarray.DataArray
        Time series of the background magnetic field.

    Returns
    -------
    p_anis : xarray.DataArray
        Time series of the pressure anisotropy.

    See also
    --------
    pyrfu.mms.rotate_tensor : Rotates pressure or temperature tensor
                                into another coordinate system.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field, ion/electron temperature and number density

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> p_xyz_i = mms.get_data("Pi_gse_fpi_fast_l2", tint, mms_id)

    Transform pressure tensor to field aligned coordinates
    >>> p_fac_i = mms.rotate_tensor(p_xyz_i, "fac", b_xyz)

    Compute pressure anistropy

    >>> p_anis = pyrf.pres_anis(p_xyz_i, b_xyz)

    """

    # Get parallel and perpendicular pressure
    p_para = p_fac.data[:, 0, 0]
    p_perp = (p_fac.data[:, 1, 1] + p_fac.data[:, 2, 2]) / 2

    # Compute magnetic pressure
    b_xyz = resample(b_xyz, p_fac)
    b_mag = np.linalg.norm(b_xyz.data, axis=1)
    p_mag = 1e-18 * b_mag**2 / (2 * constants.mu_0)

    # Compute pressure anistropy
    p_anis = (1e-9 * (p_para - p_perp) / 2) / p_mag
    p_anis = ts_scalar(p_fac.time.data, p_anis)

    return p_anis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging
import os
import warnings

# 3rd party imports
import numba
import numpy as np
import xarray as xr
from scipy import fft

from .calc_fs import calc_fs
from .cart2sph import cart2sph
from .convert_fac import convert_fac
from .resample import resample

# Local imports
from .ts_time import ts_time
from .ts_vec_xyz import ts_vec_xyz
from .unix2datetime64 import unix2datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _checksampling(e_xyz, db_xyz, b_xyz, b_bgd, flag_no_resamp):
    assert e_xyz is not None

    fs_e, fs_b = [calc_fs(e_xyz), calc_fs(db_xyz)]

    resample_b_options = {"f_s": fs_b}

    if flag_no_resamp:
        assert fs_e == fs_b
        fs_ = fs_e
    else:
        if fs_b > 1.5 * fs_e:
            e_xyz = resample(e_xyz, db_xyz, **resample_b_options)
            b_bgd = resample(b_bgd, db_xyz, **resample_b_options)

            fs_ = fs_b
            logging.info("Interpolating e to b")
        elif fs_e > 1.5 * fs_b:
            db_xyz = resample(db_xyz, e_xyz)
            b_bgd = resample(b_bgd, e_xyz)

            fs_ = fs_e
            logging.info("Interpolating b to e")
        elif fs_e == fs_b and len(e_xyz) == len(db_xyz):
            fs_ = fs_e
        else:
            fs_ = 2 * fs_e
            start_time = np.max(
                [
                    e_xyz.time.data[0].astype(np.float64) / 1e9,
                    db_xyz.time.data[0].astype(np.float64) / 1e9,
                ]
            )
            end_time = np.min(
                [
                    e_xyz.time.data[-1].astype(np.float64) / 1e9,
                    db_xyz.time.data[-1].astype(np.float64) / 1e9,
                ]
            )

            nt = np.floor((end_time - start_time) * fs_).astype(np.int64)

            t = np.linspace(start_time, end_time, nt)

            t = ts_time(t)

            e_xyz = resample(e_xyz, t)
            b_bgd = resample(b_bgd, t)
            b_xyz = resample(b_xyz, t)
            db_xyz = resample(db_xyz, t)

            logging.info("Interpolating b and e to 2x e sampling")

    return e_xyz, db_xyz, b_xyz, b_bgd, fs_


def _b_elevation(b_x, b_y, b_z, angle_b_elevation_max):
    # Remove the last sample if the total number of samples is odd
    b_x = b_x[: int(2 * (len(b_x) // 2))]
    b_y = b_y[: int(2 * (len(b_y) // 2))]
    b_z = b_z[: int(2 * (len(b_z) // 2))]

    angle_b_elevation = np.arctan(b_z / np.sqrt(b_x**2 + b_y**2))
    angle_b_elevation = np.rad2deg(angle_b_elevation)
    idx_b_par_spin_plane = np.abs(angle_b_elevation) < angle_b_elevation_max

    return angle_b_elevation, idx_b_par_spin_plane


def _freq_int(freq_int, delta_b):
    start_time = delta_b.time.data[0].astype(np.float64) / 1e9
    end_time = delta_b.time.data[-1].astype(np.float64) / 1e9

    pc12_range, other_range = [False, False]

    if isinstance(freq_int, str):
        if freq_int.lower() == "pc35":
            freq_int = [0.002, 0.1]

            delta_t = 60  # local
        else:
            pc12_range = True

            freq_int = [0.1, 5.0]

            delta_t = 1  # local

        fs_out = 1 / delta_t
    else:
        if freq_int[1] >= freq_int[0]:
            other_range = True

            fs_out = freq_int[1] / 5

            delta_t = 1 / fs_out  # local
        else:
            raise ValueError("FREQ_INT must be [f_min f_max], f_min<f_max")

    nt = np.floor((end_time - start_time) / delta_t).astype(np.int64)

    out_time = np.linspace(start_time, end_time, nt, dtype=np.float64)
    out_time += delta_t / 2.0
    out_time = out_time[:-1]

    any_range = [pc12_range, other_range]

    return any_range, freq_int, fs_out, out_time


@numba.jit(cache=True, nogil=True, parallel=True, nopython=True, fastmath=True)
def _average_data(data=None, x=None, y=None, av_window=None):
    # average data with time x to time y using window

    dtx, dty = [np.median(np.diff(x)), np.median(np.diff(y))]

    if av_window is None:
        av_window = dty

    dt2 = av_window / 2

    n_data_out = len(y)

    # Pad data with NaNs from each side
    n_point_to_add = int(np.ceil(dt2 / dtx))
    pad_nan = np.ones((n_point_to_add, data.shape[1]), dtype="complex128") * np.nan
    data_padded = np.vstack((pad_nan, data, pad_nan))

    x_pad_pref = np.linspace(x[0] - dtx * (n_point_to_add - 1), x[0], n_point_to_add)
    x_pad_suff = np.linspace(x[-1], x[-1] + dtx * (n_point_to_add - 1), n_point_to_add)
    x_padded = np.hstack((x_pad_pref, x, x_pad_suff))

    out = np.zeros((n_data_out, data.shape[1]), dtype="complex128")

    il = np.digitize(y - dt2, x_padded)
    ir = np.digitize(y + dt2, x_padded)

    for i in numba.prange(len(y)):
        for j in range(data.shape[1]):
            out[i, j] = np.nanmean(data_padded[il[i] : ir[i], j])

    return out


def _bb_xxyyzzss(power_bx_plot, power_by_plot, power_bz_plot, power_2b_plot):
    bb_xxyyzzss = np.tile(power_bx_plot[:, :, np.newaxis], (1, 1, 4))
    bb_xxyyzzss[:, :, 1] = power_by_plot
    bb_xxyyzzss[:, :, 2] = power_bz_plot
    bb_xxyyzzss[:, :, 3] = power_2b_plot
    return np.real(bb_xxyyzzss)


def _ee_xxyyzzss(power_ex_plot, power_ey_plot, power_ez_plot, power_2e_plot):
    ee_xxyyzzss = np.tile(power_ex_plot[..., np.newaxis], (1, 1, 4))
    ee_xxyyzzss[:, :, 1] = power_ey_plot
    ee_xxyyzzss[:, :, 2] = power_ez_plot
    ee_xxyyzzss[:, :, 3] = power_2e_plot
    return np.real(ee_xxyyzzss)


@numba.jit(cache=True, nogil=True, parallel=True, nopython=True, fastmath=True)
def _censure_plot(inp, idx_nan, censure, n_data, a_):
    out = inp.copy()

    for i in numba.prange(len(idx_nan) - 1):
        for j in range(len(a_)):
            if idx_nan[i] < idx_nan[i + 1]:
                out[int(max([i - censure[j], 0])) : i, j] = np.nan

            if idx_nan[i] > idx_nan[i + 1]:
                out[i : int(min([i + censure[j], n_data])), j] = np.nan

    return out


def ebsp(e_xyz, db_xyz, b_xyz, b_bgd, xyz, freq_int, **kwargs):
    """Calculates wavelet spectra of E&B and Poynting flux using wavelets
    (Morlet wavelet). Also computes polarization parameters of B using SVD
    [7]_. SVD is performed on spectral matrices computed from the time series
    of B using wavelets and then averaged over a number of wave periods.

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the wave electric field.
    db_xyz : xarray.DataArray
        Time series of the wave magnetic field.
    b_xyz : xarray.DataArray
        Time series of the high resolution background magnetic field used
        for E.B=0.
    b_bgd : xarray.DataArray
        Time series of the background magnetic field used for field aligned
        coordinates.
    xyz : xarray.DataArray
        Time series of the position time series of spacecraft used for field
        aligned coordinates.
    freq_int : str or array_like
        Frequency interval :
            * "pc12" : [0.1, 5.0]
            * "pc35" : [2e-3, 0.1]
            * [fmin, fmax] : arbitrary interval [fmin,fmax]

    Returns
    -------
    res : xarray.Dataset
        Dataset with :
            * t : xarray.DataArray
                Time.
            * f : xarray.DataArray
                Frequencies.
            * bb_xxyyzzss : xarray.DataArray
                delta_b power spectrum with :
                    * [...,0] : x
                    * [...,1] : y
                    * [...,2] : z
                    * [...,3] : sum
            * ee_xxyyzzss : xarray.DataArray
                E power spectrum with :
                    * [...,0] : x
                    * [...,1] : y
                    * [...,2] : z
                    * [...,3] : sum
            * ee_ss : xarray.DataArray
                E power spectrum (xx+yy spacecraft coordinates, e.g. ISR2).
            * pf_xyz : xarray.DataArray
                Poynting flux (xyz).
            * pf_rtp : xarray.DataArray
                Poynting flux (r, theta, phi) [angles in degrees].
            * dop : xarray.DataArray
                3D degree of polarization.
            * dop2d : xarray.DataArray
                2D degree of polarization in the polarization plane.
            * planarity : xarray.DataArray
                Planarity of polarization.
            * ellipticity : xarray.DataArray
                Ellipticity of polarization ellipse.
            * k : xarray.DataArray
                k-vector (theta, phi FAC) [angles in degrees].

    Other Parameters
    ----------------
    polarization : bool
        Computes polarization parameters. Default False.
    no_resample : bool
        No resampling, E and delta_b are given at the same time line.
        Default False.
    fac : bool
        Uses FAC coordinate system (defined by b0 and optionally xyz),
        otherwise no coordinate system transformation is performed. Default
        True.
    de_dot_b0 : bool
        Computes dEz from delta_b dot B = 0, uses full_b. Default False.
    full_b_db : bool
        delta_b contains DC field. Default False.
    nav : int
        Number of wave periods to average Default 8.
    fac_matrix : numpy.ndarray
        Specify rotation matrix to FAC system Default None.
    m_width_coeff : int or float
        Specify coefficient to multiple Morlet wavelet width by. Default 1.

    See also
    --------
    pyrfu.plot.pl_ebsp : to fill.
    pyrfu.pyrf.convert_fac : Transforms to a field-aligned coordinate.

    Notes
    -----
    This software was developed as part of the MAARBLE (Monitoring, Analyzing
    and Assessing Radiation Belt Energization and Loss) collaborative
    research project which has received funding from the European
    Community's Seventh Framework Programme (FP7-SPACE-2011-1) under grant
    agreement n. 284520.

    References
    ----------
    .. [7] 	Santolík, O., Parrot. M., and  Lefeuvre. F. (2003) Singular value
            decomposition methods for wave propagation analysis,Radio Sci.,
            38(1), 1010, doi : https://doi.org/10.1029/2000RS002523 .

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint_brst = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]
    >>> # Spacecraft index
    >>> mms_id = 3
    >>> # Load spacecraft position
    >>> tint_long = pyrf.extend_tint(tint_brst, [-100, 100])
    >>> r_xyz = mms.get_data("R_gse", tint_long, mms_id)
    >>> # Load background magnetic field, electric field and magnetic field
    fluctuations
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint_brst, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint_brst, mms_id)
    >>> b_scm = mms.get_data("B_gse_scm_brst_l2", tint_brst, mms_id)
    >>> # Polarization analysis
    >>> options = dict(polarization=True, fac=True)
    >>> polarization = pyrf.ebsp(e_xyz, b_scm, b_xyz, b_xyz, r_xyz,
    >>>                          freq_int=[10, 4000], **options)

    """

    assert isinstance(db_xyz, xr.DataArray), "delta_b must be a DataArray"
    assert isinstance(b_xyz, xr.DataArray), "full_b must be a DataArray"
    assert isinstance(b_bgd, xr.DataArray), "b0 must be a DataArray"

    message = "freq_int must be a string or array_like"
    assert isinstance(freq_int, (list, np.ndarray, str)), message

    if isinstance(freq_int, (list, np.ndarray)):
        assert len(freq_int) == 2, "freq_int list must contain two elements"
    else:
        assert freq_int in ["pc12", "pc35"], "string freq_int must be pc12 or pc35"

    # Compute magnetic field fluctuations sampling frequency
    fs_b = calc_fs(db_xyz)

    # Below which we cannot apply E*B=0
    angle_b_elevation_max = 15.0

    want_ee = e_xyz is not None

    res = {
        "t": None,
        "f": None,
        "flagFac": 0,
        "bb_xxyyzzss": None,
        "ee_xxyyzzss": None,
        "ee_ss": None,
        "pf_xyz": None,
        "pf_rtp": None,
        "dop": None,
        "dop2d": None,
        "planarity": None,
        "ellipticity": None,
        "k_tp": None,
        "full_b": b_xyz,
        "b0": b_bgd,
        "r": xyz,
    }

    want_polarization = kwargs.get("polarization", False)

    flag_no_resample = kwargs.get("no_resample", False)
    flag_want_fac = kwargs.get("fac", True)
    flag_de_dot_b0 = kwargs.get("de_dot_b0", False)
    flag_full_b_db = kwargs.get("full_b_db", False)

    m_width_coeff = kwargs.get("m_width_coeff", 1.0)

    # Number of wave periods to average
    n_wave_period_to_average = kwargs.get("nav", 8)

    # matrix for rotation to FAC
    fac_matrix = kwargs.get("fac_matrix", None)

    if flag_want_fac and fac_matrix is None:
        if xyz is None:
            logging.info(
                "convert_fac : assuming s/c position [1 0 0] for estimating FAC"
            )
            xyz = [1, 0, 0]
            xyz = ts_vec_xyz(db_xyz.time.data, np.tile(xyz, (len(db_xyz), 1)))
        else:
            assert isinstance(xyz, xr.DataArray), "xyz must be a DataArray"

        xyz = resample(xyz, db_xyz, **{"f_s": fs_b})

    b_bgd = resample(b_bgd, db_xyz, **{"f_s": fs_b})

    if flag_full_b_db:
        b_xyz = db_xyz
        res["full_b"] = b_xyz
        db_xyz = db_xyz - b_bgd

    any_range, freq_int, out_sampling, out_time = _freq_int(freq_int, db_xyz)
    pc12_range, other_range = any_range

    if want_ee:
        # Check the sampling rate
        temp_ = _checksampling(e_xyz, db_xyz, b_xyz, b_bgd, flag_no_resample)
        e_xyz, db_xyz, b_xyz, b_bgd, in_sampling = temp_

    else:
        in_sampling = calc_fs(db_xyz)

        e_xyz = None

    if in_sampling / 2 < freq_int[1]:
        raise ValueError("F_MAX must be lower than the Nyquist frequency")

    if want_ee and e_xyz.shape[1] < 3 and not flag_de_dot_b0:
        raise TypeError(
            "E must have all 3 components or flag de_dot_db=0 must be given"
        )

    if len(db_xyz) % 2:
        db_xyz = db_xyz[:-1, :]
        b_bgd = b_bgd[:-1, :]

        if fac_matrix is None:
            xyz = xyz[:-1, :]
        else:
            fac_matrix = fac_matrix[:-1, ...]

        if want_ee:
            e_xyz = e_xyz[:-1, :]

    in_time = db_xyz.time.data.astype(np.float64) / 1e9

    b_x, b_y, b_z = [None, None, None]

    idx_b_par_spin_plane = None

    if flag_de_dot_b0:
        b_x, b_y, b_z = [b_xyz[:, i].data for i in range(3)]

        # Remove the last sample if the total number of samples is odd
        # temp_ = _b_elevation(b_x, b_y, b_z, angle_b_elevation_max)
        # angle_b_elevation, idx_b_par_spin_plane = temp_
        _, idx_b_par_spin_plane = _b_elevation(b_x, b_y, b_z, angle_b_elevation_max)

    # If E has all three components, transform E and B waveforms to a magnetic
    # field aligned coordinate (FAC) and save eisr for computation of e_sum.
    # Otherwise we compute Ez within the main loop and do the transformation
    # to FAC there.

    time_b0 = 0
    if flag_want_fac:
        res["flagFac"] = True

        time_b0 = b_bgd.time.data.astype(np.float64) / 1e9

        if want_ee and not flag_de_dot_b0:
            eisr2 = e_xyz[:, :2]
            idx_nan_e = np.isnan(e_xyz.data)
            idx_nan_eisr2 = np.isnan(eisr2.data)

            if fac_matrix is None:
                e_xyz = convert_fac(e_xyz, b_bgd, xyz)
            else:
                e_xyz = convert_fac(e_xyz, fac_matrix)

        else:
            idx_nan_e = np.full((len(in_time), 3), False)
            eisr2 = None
            idx_nan_eisr2 = np.full((len(in_time), 2), False)

        if fac_matrix is None:
            db_xyz = convert_fac(db_xyz, b_bgd, xyz)
        else:
            db_xyz = convert_fac(db_xyz, fac_matrix)
    else:
        idx_nan_e = np.full((len(in_time), 3), False)
        eisr2 = None
        idx_nan_eisr2 = np.full((len(in_time), 2), False)

    # Find the frequencies for an FFT of all data and set important parameters
    nd2 = len(in_time) / 2

    freq = in_sampling * np.arange(nd2) / nd2 * 0.5

    # The frequencies corresponding to FFT
    w_ = np.hstack([0, freq, -np.flip(freq[:-1])])

    morlet_width = 5.36 * m_width_coeff

    # to get proper overlap for Morlet
    freq_number = np.ceil(
        (np.log10(freq_int[1]) - np.log10(freq_int[0])) * 12 * m_width_coeff
    )

    a_min, a_max = [
        np.log10(0.5 * in_sampling / freq_int[1]),
        np.log10(0.5 * in_sampling / freq_int[0]),
    ]

    a_number = freq_number
    a_ = np.logspace(a_min, a_max, int(a_number))

    # Maximum frequency
    w_0 = in_sampling / 2

    # Width of the Morlet wavelet
    sigma = morlet_width / w_0

    # Make the FFT of all data
    idx_nan_b = np.isnan(db_xyz.data)

    db_xyz.data[idx_nan_b] = 0

    swb = fft.fft(db_xyz.data, axis=0, workers=os.cpu_count())

    sw_e, sw_eisr2 = [None, None]

    if want_ee:
        logging.info("ebsp ... calculate E and B wavelet transform ... ")
        e_xyz.data[idx_nan_e] = 0.0

        sw_e = fft.fft(e_xyz.data, axis=0, workers=os.cpu_count())

        if flag_want_fac and not flag_de_dot_b0:
            eisr2.data[idx_nan_eisr2] = 0.0

            sw_eisr2 = fft.fft(eisr2.data, axis=0, workers=os.cpu_count())
    else:
        logging.info("ebsp ... calculate B wavelet transform ....")

    # Loop through all frequencies
    n_data, n_freq, n_data_out = [len(in_time), len(a_), len(out_time)]

    #
    power_ex_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_ey_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_ez_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_2e_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_2e_isr2_plot = np.zeros((n_data, n_freq), dtype="complex128")

    power_bx_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_by_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_bz_plot = np.zeros((n_data, n_freq), dtype="complex128")
    power_2b_plot = np.zeros((n_data, n_freq), dtype="complex128")

    s_plot_x = np.zeros((n_data, n_freq))
    s_plot_y = np.zeros((n_data, n_freq))
    s_plot_z = np.zeros((n_data, n_freq))

    planarity, ellipticity = [np.zeros((n_data_out, n_freq)) for _ in range(2)]
    dop_3d = np.zeros((n_data_out, n_freq), dtype="complex128")
    dop_2d = np.zeros((n_data_out, n_freq), dtype="complex128")

    the_svd_fac = np.zeros((n_data_out, n_freq))
    phi_svd_fac = np.zeros((n_data_out, n_freq))

    # Get the correct frequencies for the wavelet transform
    frequency_vec = w_0 / a_

    censure = np.floor(2 * a_ * out_sampling / in_sampling * n_wave_period_to_average)

    for ind_a, a_0 in enumerate(a_):
        new_freq_mat = w_0 / a_0

        # resample to 1 second sampling for Pc1-2 or 1 minute sampling for
        # Pc3-5 average top frequencies to 1 second/1 minute below will be
        # an average over 8 wave periods. first find where one sample is less
        # than eight wave periods

        if frequency_vec[ind_a] / n_wave_period_to_average > out_sampling:
            av_window = 1 / out_sampling
        else:
            av_window = n_wave_period_to_average / frequency_vec[ind_a]

        # Get the wavelet transform by backward FFT
        w_exp_mat = np.exp(-sigma * sigma * ((a_0 * w_ - w_0) ** 2) / 2)
        w_exp_mat2 = np.tile(w_exp_mat[:, np.newaxis], (1, 2))
        w_exp_mat = np.tile(w_exp_mat[:, np.newaxis], (1, 3))

        wb = fft.ifft(np.sqrt(1) * swb * w_exp_mat, axis=0, workers=os.cpu_count())
        wb = np.array(wb)  # Make sure it's an array (scipy.fft.ifft returns Any type)
        wb[idx_nan_b] = np.nan

        we, w_eisr2 = [None, None]

        if want_ee:
            we = fft.ifft(np.sqrt(1) * sw_e * w_exp_mat, axis=0, workers=os.cpu_count())
            we = np.array(we)
            we[idx_nan_e] = np.nan

            if flag_want_fac and not flag_de_dot_b0:
                w_eisr2 = fft.ifft(
                    np.sqrt(1) * sw_eisr2 * w_exp_mat2, axis=0, workers=os.cpu_count()
                )
                w_eisr2 = np.array(w_eisr2)
                w_eisr2[idx_nan_eisr2] = np.nan

                # Power spectrum of E, power = (2*pi)*conj(W).*W./new_freq_mat
                power_2e_isr2_plot[:, ind_a] = np.sum(
                    2 * np.pi * (w_eisr2 * np.conj(w_eisr2)) / new_freq_mat, axis=1
                )
            else:
                # Power spectrum of E, power = (2*pi)*conj(W).*W./new_freq_mat
                power_2e_isr2_plot[:, ind_a] = np.sum(
                    2 * np.pi * (we * np.conj(we)) / new_freq_mat, axis=1
                )

            # Compute Ez from dE * B = 0
            if flag_de_dot_b0:
                we_re, we_im = [np.real(we), np.imag(we)]

                we_z = (
                    -(we_re[:, 0] * b_x + we_re[:, 1] * b_y) / b_z
                    - 1j * (we_im[:, 0] * b_x + we_im[:, 1] * b_y) / b_z
                )
                we_z[idx_b_par_spin_plane] = np.nan

                if flag_want_fac:
                    if fac_matrix is None:
                        tmp = np.vstack([np.transpose(we[:, :2]), np.transpose(we_z)])
                        arg_ = ts_vec_xyz(time_b0, np.transpose(tmp))
                        we = convert_fac(arg_, b_bgd, xyz)
                    else:
                        tmp = np.vstack([np.transpose(we[:, :2]), np.transpose(we_z)])
                        arg_ = ts_vec_xyz(time_b0, np.transpose(tmp))
                        we = convert_fac(arg_, fac_matrix)
                else:
                    we = np.transpose(
                        np.vstack([np.transpose(we[:, :2]), np.transpose(we_z)])
                    )

            power_e = 2 * np.pi * (we * np.conj(we)) / new_freq_mat
            power_e = np.vstack([power_e.T, np.sum(power_e, axis=1)]).T

            power_ex_plot[:, ind_a] = power_e[:, 0]
            power_ey_plot[:, ind_a] = power_e[:, 1]
            power_ez_plot[:, ind_a] = power_e[:, 2]
            power_2e_plot[:, ind_a] = power_e[:, 3]

            # Poynting flux calculations, assume E and b units mV/m and nT,
            # get  S in uW/m^2 4pi from wavelets, see A. Tjulins power
            # estimates
            coeff_poynting = 10 / 4 / np.pi * (1 / 4) * (4 * np.pi)

            s = np.zeros((n_data, 3))

            we_x, we_y, we_z = [we[:, i] for i in range(3)]
            wb_x, wb_y, wb_z = [wb[:, i] for i in range(3)]

            s[:, 0] = (
                coeff_poynting
                * np.real(
                    we_y * np.conj(wb_z)
                    + np.conj(we_y) * wb_z
                    - we_z * np.conj(wb_y)
                    - np.conj(we_z) * wb_y
                )
                / new_freq_mat
            )
            s[:, 1] = (
                coeff_poynting
                * np.real(
                    we_z * np.conj(wb_x)
                    + np.conj(we_z) * wb_x
                    - we_x * np.conj(wb_z)
                    - np.conj(we_x) * wb_z
                )
                / new_freq_mat
            )
            s[:, 2] = (
                coeff_poynting
                * np.real(
                    we_x * np.conj(wb_y)
                    + np.conj(we_x) * wb_y
                    - we_y * np.conj(wb_x)
                    - np.conj(we_y) * wb_x
                )
                / new_freq_mat
            )

            s_plot_x[:, ind_a] = s[:, 0]
            s_plot_y[:, ind_a] = s[:, 1]
            s_plot_z[:, ind_a] = s[:, 2]

        # Power spectrum of B
        power_b = 2 * np.pi * (wb * np.conj(wb)) / new_freq_mat
        power_b = np.vstack([power_b.T, np.sum(power_b, axis=1)]).T

        power_bx_plot[:, ind_a] = power_b[:, 0]
        power_by_plot[:, ind_a] = power_b[:, 1]
        power_bz_plot[:, ind_a] = power_b[:, 2]
        power_2b_plot[:, ind_a] = power_b[:, 3]

        # Polarization parameters
        if want_polarization:
            # Construct spectral matrix and average it
            s_mat = np.zeros((n_data, 3, 3), dtype="complex128")

            for i in range(3):
                for j in range(3):
                    s_mat[:, i, j] = (
                        2 * np.pi * (wb[:, i] * np.conj(wb[:, j])) / new_freq_mat
                    )

            # Averaged s_mat
            s_mat_avg = np.zeros((n_data_out, 3, 3), dtype="complex128")

            for comp in range(3):
                s_mat_avg[..., comp] = _average_data(
                    s_mat[..., comp], in_time, out_time, av_window
                )

            # Remove data possibly influenced by edge effects
            censure_idx = np.hstack(
                [
                    np.arange(np.min([censure[ind_a], len(out_time)])),
                    np.arange(
                        np.max([0, len(out_time) - censure[ind_a] - 1]), len(out_time)
                    ),
                ]
            )
            censure_idx = censure_idx.astype(np.int64)

            s_mat_avg[censure_idx, ...] = np.nan

            # compute singular value decomposition
            # real matrix which is superposition of real part of spectral
            # matrix over imaginary part
            a_mat, u_mat = [np.zeros((6, 3, n_data_out)) for _ in range(2)]
            w_mat, v_mat = [np.zeros((3, 3, n_data_out)) for _ in range(2)]

            # wSingularValues = zeros(3,n_data2);
            # R = zeros(3,3,n_data2); #spectral matrix in coordinate defined
            # by V axes
            a_mat[:3, ...] = np.real(np.transpose(s_mat_avg, [1, 2, 0]))
            a_mat[3:6, ...] = -np.imag(np.transpose(s_mat_avg, [1, 2, 0]))

            for i in range(n_data_out):
                if np.sum(np.isnan(a_mat[..., i])) != 0:
                    u_mat[..., i] = np.nan
                    w_mat[..., i] = np.nan
                    v_mat[..., i] = np.nan
                else:
                    uu_, ww_, vv_ = np.linalg.svd(a_mat[..., i], full_matrices=False)
                    u_mat[..., i] = uu_
                    w_mat[..., i] = ww_
                    v_mat[..., i] = vv_

            # compute direction of propagation
            sign_kz = np.sign(v_mat[2, 2, :])
            v_mat[2, 2, :] = v_mat[2, 2, :] * sign_kz
            v_mat[1, 2, :] = v_mat[1, 2, :] * sign_kz
            v_mat[0, 2, :] = v_mat[0, 2, :] * sign_kz

            the_svd_fac[:, ind_a] = np.abs(
                np.squeeze(
                    np.arctan(
                        np.sqrt(v_mat[0, 2, :] ** 2 + v_mat[1, 2, :] ** 2)
                        / v_mat[2, 2, :]
                    )
                )
            )
            phi_svd_fac[:, ind_a] = np.squeeze(
                np.arctan2(v_mat[1, 2, :], v_mat[0, 2, :])
            )

            # Calculate polarization parameters
            planarity_local = np.squeeze(1 - np.sqrt(w_mat[2, 2, :] / w_mat[0, 0, :]))
            planarity_local[censure_idx] = np.nan

            planarity[:, ind_a] = planarity_local

            # ellipticity: ratio of axes of polarization ellipse axes*sign of
            # polarization

            ellipticity_local = np.squeeze(w_mat[1, 1, :] / w_mat[0, 0, :]) * np.sign(
                np.imag(s_mat_avg[:, 0, 1])
            )
            ellipticity_local[censure_idx] = np.nan

            ellipticity[:, ind_a] = ellipticity_local

            # DOP = sqrt[(3/2.*trace(SM^2)./(trace(SM))^2 - 1/2)]; Samson, 1973, JGR
            dop = np.sqrt(
                (3 / 2)
                * (
                    np.trace(np.matmul(s_mat_avg, s_mat_avg), axis1=1, axis2=2)
                    / np.trace(s_mat_avg, axis1=1, axis2=2) ** 2
                )
                - 1 / 2
            )

            dop[censure_idx] = np.nan
            dop_3d[:, ind_a] = dop

            # DOP in 2D = sqrt[2*trace(rA^2)/trace(rA)^2 - 1)]; Ulrich
            v_mat_new = np.transpose(v_mat, [2, 0, 1])

            s_mat_avg2dim = np.matmul(
                v_mat_new, np.matmul(s_mat_avg, np.transpose(v_mat_new, [0, 2, 1]))
            )
            s_mat_avg2dim = s_mat_avg2dim[:, :2, :2]
            s_mat_avg = s_mat_avg2dim

            dop2dim = np.sqrt(
                2
                * (
                    np.trace(np.matmul(s_mat_avg, s_mat_avg), axis1=1, axis2=2)
                    / np.trace(s_mat_avg, axis1=1, axis2=2) ** 2
                )
                - 1
            )
            dop2dim[censure_idx] = np.nan
            dop_2d[:, ind_a] = dop

    # set data gaps to NaN and remove edge effects
    censure = np.floor(2 * a_)

    for ind_a in range(len(a_)):
        censure_idx = np.hstack(
            [
                np.arange(np.min([censure[ind_a], len(in_time)])),
                np.arange(np.max([1, len(in_time) - censure[ind_a]]), len(in_time)),
            ]
        )

        censure_idx = censure_idx.astype(np.int64)

        power_bx_plot[censure_idx, ind_a] = np.nan
        power_by_plot[censure_idx, ind_a] = np.nan
        power_bz_plot[censure_idx, ind_a] = np.nan
        power_2b_plot[censure_idx, ind_a] = np.nan

        if want_ee:
            power_ex_plot[censure_idx, ind_a] = np.nan
            power_ey_plot[censure_idx, ind_a] = np.nan
            power_ez_plot[censure_idx, ind_a] = np.nan
            power_2e_plot[censure_idx, ind_a] = np.nan

            power_2e_isr2_plot[censure_idx, ind_a] = np.nan

            s_plot_x[censure_idx, ind_a] = np.nan
            s_plot_y[censure_idx, ind_a] = np.nan
            s_plot_z[censure_idx, ind_a] = np.nan

    # remove edge effects from data gaps
    idx_nan_e = np.sum(idx_nan_e, axis=1) > 0
    idx_nan_b = np.sum(idx_nan_b, axis=1) > 0
    idx_nan_eisr2 = np.sum(idx_nan_eisr2, axis=1) > 0

    n_power_b = len(power_2b_plot)

    if pc12_range or other_range:
        censure3 = np.floor(1.8 * a_)
    else:
        censure3 = np.floor(0.4 * a_)

    # Censure magnetic fied
    power_bx_plot = _censure_plot(power_bx_plot, idx_nan_b, censure3, n_power_b, a_)
    power_by_plot = _censure_plot(power_by_plot, idx_nan_b, censure3, n_power_b, a_)
    power_bz_plot = _censure_plot(power_bz_plot, idx_nan_b, censure3, n_power_b, a_)
    power_2b_plot = _censure_plot(power_2b_plot, idx_nan_b, censure3, n_power_b, a_)

    # Censure electric field
    n_power_e = len(power_2e_plot)
    power_ex_plot = _censure_plot(power_ex_plot, idx_nan_e, censure3, n_power_e, a_)
    power_ey_plot = _censure_plot(power_ey_plot, idx_nan_e, censure3, n_power_e, a_)
    power_ez_plot = _censure_plot(power_ez_plot, idx_nan_e, censure3, n_power_e, a_)
    power_2e_plot = _censure_plot(power_2e_plot, idx_nan_e, censure3, n_power_e, a_)

    power_2e_isr2_plot = _censure_plot(
        power_2e_isr2_plot, idx_nan_e, censure3, n_power_e, a_
    )

    # Censure poynting flux
    s_plot_x = _censure_plot(s_plot_x, idx_nan_b, censure3, n_power_b, a_)
    s_plot_x = _censure_plot(s_plot_x, idx_nan_e, censure3, n_power_e, a_)
    s_plot_y = _censure_plot(s_plot_y, idx_nan_b, censure3, n_power_b, a_)
    s_plot_y = _censure_plot(s_plot_y, idx_nan_e, censure3, n_power_e, a_)
    s_plot_z = _censure_plot(s_plot_z, idx_nan_b, censure3, n_power_b, a_)
    s_plot_z = _censure_plot(s_plot_z, idx_nan_e, censure3, n_power_e, a_)

    n_power_2e_isr2 = len(power_2e_isr2_plot)
    power_2e_isr2_plot = _censure_plot(
        power_2e_isr2_plot, idx_nan_eisr2, censure3, n_power_2e_isr2, a_
    )

    power_bx_plot = _average_data(power_bx_plot, in_time, out_time)
    power_by_plot = _average_data(power_by_plot, in_time, out_time)
    power_bz_plot = _average_data(power_bz_plot, in_time, out_time)
    power_2b_plot = _average_data(power_2b_plot, in_time, out_time)

    bb_xxyyzzss = _bb_xxyyzzss(
        power_bx_plot, power_by_plot, power_bz_plot, power_2b_plot
    )

    # Output
    res["t"] = unix2datetime64(out_time)
    res["f"] = frequency_vec[::-1]
    res["bb_xxyyzzss"] = xr.DataArray(
        bb_xxyyzzss[:, ::-1, ...],
        coords=[res["t"], res["f"], ["xx", "yy", "zz", "ss"]],
        dims=["time", "frequency", "comp"],
    )

    if want_ee:
        power_ex_plot = _average_data(power_ex_plot, in_time, out_time)
        power_ey_plot = _average_data(power_ey_plot, in_time, out_time)
        power_ez_plot = _average_data(power_ez_plot, in_time, out_time)
        power_2e_plot = _average_data(power_2e_plot, in_time, out_time)

        power_2e_isr2_plot = _average_data(power_2e_isr2_plot, in_time, out_time)
        power_2e_isr2_plot = np.real(power_2e_isr2_plot)

        s_plot_x = np.real(_average_data(s_plot_x, in_time, out_time))
        s_plot_y = np.real(_average_data(s_plot_y, in_time, out_time))
        s_plot_z = np.real(_average_data(s_plot_z, in_time, out_time))

        # TODO: check that it's correct (MATLAB weird stuff)
        s_azimuth, s_elevation, s_r = cart2sph(s_plot_x, s_plot_y, s_plot_z)

        ee_xxyyzzss = _ee_xxyyzzss(
            power_ex_plot, power_ey_plot, power_ez_plot, power_2e_plot
        )

        poynting_xyz = np.tile(s_plot_x, (3, 1, 1))
        poynting_xyz = np.transpose(poynting_xyz, [1, 2, 0])
        poynting_xyz[:, :, 1] = s_plot_y
        poynting_xyz[:, :, 2] = s_plot_z
        poynting_xyz = poynting_xyz.astype(np.float64)

        poynting_r_th_ph = np.tile(s_r, (3, 1, 1))
        poynting_r_th_ph = np.transpose(poynting_r_th_ph, [1, 2, 0])
        poynting_r_th_ph[..., 1] = np.pi / 2 - s_elevation
        poynting_r_th_ph[..., 2] = s_azimuth
        poynting_r_th_ph[..., 1:] = poynting_r_th_ph[..., 1:] * 180 / np.pi
        poynting_r_th_ph = poynting_r_th_ph.astype(np.float64)

        # Output
        res["ee_ss"] = power_2e_isr2_plot.astype(np.float64)

        res["ee_xxyyzzss"] = xr.DataArray(
            ee_xxyyzzss[:, ::-1, ...],
            coords=[res["t"], res["f"], ["xx", "yy", "zz", "ss"]],
            dims=["time", "frequency", "comp"],
        )

        res["pf_xyz"] = xr.DataArray(
            poynting_xyz[:, ::-1, ...],
            coords=[res["t"], res["f"], ["x", "y", "z"]],
            dims=["time", "frequency", "comp"],
        )

        res["pf_rtp"] = xr.DataArray(
            poynting_r_th_ph[:, ::-1, ...],
            coords=[res["t"], res["f"], ["rho", "theta", "phi"]],
            dims=["time", "frequency", "comp"],
        )

    if want_polarization:
        # Define parameters for which we cannot compute the wave vector
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=RuntimeWarning)
            ind_low_planarity = planarity < 0.5
            ind_low_ellipticity = np.abs(ellipticity) < 0.2

        the_svd_fac[ind_low_planarity] = np.nan
        phi_svd_fac[ind_low_planarity] = np.nan

        the_svd_fac[ind_low_ellipticity] = np.nan
        phi_svd_fac[ind_low_ellipticity] = np.nan

        k_th_ph_svd_fac = np.zeros((the_svd_fac.shape[0], the_svd_fac.shape[1], 2))
        k_th_ph_svd_fac[..., 0] = the_svd_fac
        k_th_ph_svd_fac[..., 1] = phi_svd_fac

        # Output
        res["dop"] = xr.DataArray(
            np.real(dop_3d[:, ::-1]),
            coords=[res["t"], res["f"]],
            dims=["time", "frequency"],
        )

        res["dop2d"] = xr.DataArray(
            np.real(dop_2d[:, ::-1]),
            coords=[res["t"], res["f"]],
            dims=["time", "frequency"],
        )

        res["planarity"] = xr.DataArray(
            planarity[:, ::-1], coords=[res["t"], res["f"]], dims=["time", "frequency"]
        )

        res["ellipticity"] = xr.DataArray(
            ellipticity[:, ::-1],
            coords=[res["t"], res["f"]],
            dims=["time", "frequency"],
        )

        res["k_tp"] = xr.DataArray(
            k_th_ph_svd_fac[:, ::-1, ...],
            coords=[res["t"], res["f"], ["theta", "phi"]],
            dims=["time", "frequency", "comp"],
        )

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import random
import unittest

# 3rd party imports
import numpy as np
import xarray as xr
from ddt import data, ddt

# Local imports
from .. import dispersion

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.4"
__status__ = "Prototype"


class DispSurfCalcTestCase(unittest.TestCase):
    def test_disp_surf_calc_output(self):
        kx, kz, wf, extra_param = dispersion.disp_surf_calc(
            random.random(), random.random(), random.random(), random.random()
        )
        self.assertIsInstance(kx, np.ndarray)
        self.assertIsInstance(kz, np.ndarray)
        self.assertIsInstance(wf, np.ndarray)
        self.assertIsInstance(extra_param, dict)


@ddt
class OneFluidDispersionTestCase(unittest.TestCase):
    @data(
        (
            random.random(),
            random.random(),
            {"n": random.random(), "t": random.random(), "gamma": random.random()},
            {"n": random.random(), "t": random.random(), "gamma": random.random()},
            random.randint(10, 1000),
        )
    )
    def test_one_fluid_dispersion_output(self, value):
        b_0, theta, ions, electrons, n_k = value
        result = dispersion.one_fluid_dispersion(b_0, theta, ions, electrons, n_k)
        self.assertIsInstance(result[0], xr.DataArray)
        self.assertIsInstance(result[1], xr.DataArray)
        self.assertIsInstance(result[2], xr.DataArray)


if __name__ == "__main__":
    unittest.main()



# Built-in imports
import os
import sys

# 3rd party imports
import pydata_sphinx_theme

# -- Path setup --------------------------------------------------------------
src_path = os.path.abspath("..")
sys.path.insert(0, src_path)
os.environ["PYTHONPATH"] = src_path

# -- General configuration ---------------------------------------------------
autosummary_generate = True

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_gallery.load_style",
    "sphinx_codeautolink",
    "sphinxcontrib.apidoc",
    "sphinx.ext.todo",
    "nbsphinx",
    "sphinx_copybutton",
    "numpydoc",
]


apidoc_module_dir = "../pyrfu"
apidoc_output_dir = "dev"
apidoc_excluded_paths = ["tests", "*/tests/", "*/*/tests/", "*/*/*/tests"]
apidoc_separate_modules = True
apidoc_module_first = True

# autosectionlabel_prefix_document = True
codeautolink_custom_blocks = {
    "python3": None,
    "pycon3": "sphinx_codeautolink.clean_pycon",
}

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc=figure.dpi=96",
]

nbsphinx_execute = "never"

# nbsphinx_thumbnails = {}

mathjax3_config = {
    "tex": {"tags": "ams", "useLabelIds": True},
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "pyrfu"
author = "Louis Richard"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "examples/**/README.md",
]

numpydoc_show_class_members = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'sphinx_rtd_theme'
html_theme = pydata_sphinx_theme.__name__

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don"t match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``["localtoc.html", "relations.html", "sourcelink.html",
# "searchbox.html"]``.
#
# html_sidebars = {}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
html_logo = "_static/logo-pyrfu.png"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or
# 32x32 pixels large.

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pyrfudoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements: dict[str, str] = {
    "papersize": "letterpaper",  # The paper size ('letterpaper' or 'a4paper').
    "pointsize": "10pt",  # The font size ('10pt', '11pt' or '12pt').
    "preamble": "",  # Additional stuff for the LaTeX preamble.
    "figure_align": "htbp",  # Latex figure (float) alignment
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "pyrfu.tex",
        "Python Space Physics (RymdFysik) Utilities Documentation",
        "Louis Richard",
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "pyrfu",
        "Python Space Physics (RymdFysik) Utilities Documentation",
        [author],
        1,
    )
]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "pyrfu",
        "Python Space Physics (RymdFysik) Utilities Documentation",
        author,
        "pyrfu",
        "Python routines to work with space data, particularly with MMS data. "
        "Also some general plasma routines.",
        "Miscellaneous",
    ),
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ""

# A unique identification for the text.
#
# epub_uid = ""

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "xarray": ("http://xarray.pydata.org/en/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
}

autodoc_mock_imports = ["numba", "sfs"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def waverage(inp, f_sampl: float = None, n_pts: int = 7):
    r"""Computes weighted average.

    Parameters
    ----------
    inp : array_like
        Input data
    f_sampl : float, Optional
        Sampling frequency.
    n_pts : int, Optional
        Number of point to average over 5 ot 7. Default is 7

    Returns
    -------
    out : ndarray
        Weighted averaged of inp

    """

    assert n_pts in [5, 7], "n_pts  must be 5 or 7"

    if f_sampl is None:
        f_sampl = 1e9 / (inp.time.data[1] - inp.time.data[0]).view("i8")

    n_data = np.round(
        1e-9 * (inp.time.data[-1] - inp.time.data[0]).view("i8") * f_sampl,
    )

    inp_data = inp.data
    try:
        n_cols = inp_data.shape[1]
    except IndexError:
        inp_data = inp_data[:, None]
        n_cols = inp_data.shape[1]

    f_sampl = 1e9 * n_data / (inp.time.data[-1] - inp.time.data[0]).view("i8")
    delta_t = 1 / f_sampl

    out = np.zeros((n_data + 1, n_cols))
    out[:, 0] = np.linspace(inp_data[0, 0], inp_data[-1, 0], n_data + 1)
    indices = np.round((inp_data[:, 0] - inp_data[0, 0]) / delta_t + 1)
    out[indices, :] = inp_data[:, 1:]
    out[np.isnan(out)] = 0  # set NaNs to zeros

    for col in range(n_cols):
        if n_pts == 5:
            new_data = np.hstack([0, 0, out[:col], 0, 0])
        elif n_pts == 7:
            new_data = np.hstack([0, 0, 0, out[:col], 0, 0, 0])
        else:
            raise ValueError("n_pts must be 5 or 7")

        for j in range(n_data + 1):
            out[j, col] = _wave(new_data[j : j + n_pts - 1], n_pts)

    # Make sure we do return matrix of the same size
    out = out[indices, :]

    return out


def _wave(inp_window, n_pts):
    """computes weighted average"""

    if n_pts == 5:
        m = [0.1, 0.25, 0.3, 0.25, 0.1]
    elif n_pts == 7:
        m = [0.07, 0.15, 0.18, 0.2, 0.18, 0.15, 0.07]
    else:
        raise ValueError("n_pts must be 5 or 7")

    # find missing points == 0
    if np.sum(m[inp_window == 0]) == 1:
        average = 0
    else:
        average = np.sum(inp_window * m) / (1 - np.sum(m[inp_window == 0]))

    return average



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def vdf_omni(vdf, method: str = "mean"):
    r"""Computes omni-directional distribution, without changing the units.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the 3D velocity distribution with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    method : {"mean", "sum"}, Optional
        Method of computation. Use "sum" for counts and "mean" for
        everything else. Default is "mean".

    Returns
    -------
    out : xarray.DataArray
        Time series of the omnidirectional velocity distribution function.

    """

    assert method.lower() in ["mean", "sum"], "invalid method!!"

    time = vdf.time.data

    energy = vdf.energy.data
    thetas = vdf.theta.data
    dangle = np.pi / vdf.theta.shape[0]
    np_phi = vdf.phi.shape[1]

    sine_theta = np.ones((np_phi, 1)) * np.sin(np.deg2rad(thetas))
    solid_angles = dangle * dangle * sine_theta
    all_solid_angles = np.tile(
        solid_angles,
        (len(time), energy.shape[1], 1, 1),
    )

    if method.lower() == "mean":
        dist = vdf.data.data * all_solid_angles
        omni = np.squeeze(np.nanmean(np.nanmean(dist, axis=3), axis=2))
        omni /= np.mean(np.mean(solid_angles))
    else:
        dist = vdf.data.data
        omni = np.squeeze(np.nansum(np.nansum(dist, axis=3), axis=2))

    energy = np.mean(energy[:2, :], axis=0)

    # Use global and zVariable attributes
    attrs = {**vdf.data.attrs, **vdf.attrs}
    attrs = {k: attrs[k] for k in sorted(attrs)}

    out = xr.DataArray(
        omni,
        coords=[time, energy],
        dims=["time", "energy"],
        attrs=attrs,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.pyplot as plt
import numpy as np

# Local imports
from pyrfu.models.ion_anisotropy_thresh import ion_anisotropy_thresh


def ion_brazil_plot_thresh(ax, legend: bool = True, colors: list = None):
    r"""Add the thresholds for the electromagnetic ion temperature anisotropy driven
    instabilities from linear Vlasov theory [1]_ [2]_ . The thresholds are given at
    :math:`\gamma = 10^{-2} \omega_{ci}` . Also changes the scale and limits of the
    axis to make it in form of a Brazil plot.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        Axis where to plot the thresholds in a Brazil plot format.


    References
    ----------
    .. [1]  Daniel Verscharen et al 2016 ApJ 831 128
    .. [2]  Bennett A. Maruca et al 2012 ApJ 748 137

    """

    if colors is None:
        colors = ["deeppink", "m", "b", "c"]

    # Proton - electron plasma thresholds [1]_
    # Proton cyclotron instability
    (l0,) = ax.loglog(
        np.logspace(-2, 4),
        ion_anisotropy_thresh(np.logspace(-2, 4), "proton-cyclotron", "10^-2"),
        color=colors[0],
        linestyle="--",
        lw=1.5,
    )

    # Mirror-mode instability
    (l1,) = ax.loglog(
        np.logspace(-2, 4),
        ion_anisotropy_thresh(np.logspace(-2, 4), "mirror", "10^-2"),
        color=colors[1],
        linestyle="--",
        lw=1.5,
    )

    # Parallel firehose
    (l2,) = ax.loglog(
        np.logspace(-2, 4),
        ion_anisotropy_thresh(np.logspace(-2, 4), "parallel-firehose", "10^-2"),
        color=colors[2],
        linestyle="--",
        lw=1.5,
    )

    # Oblique firehose
    (l3,) = ax.loglog(
        np.logspace(-2, 4),
        ion_anisotropy_thresh(np.logspace(-2, 4), "oblique-firehose", "10^-2"),
        color=colors[3],
        linestyle="--",
        lw=1.5,
    )
    ax.set_xlim([1e-2, 5e3])
    ax.set_ylim([0.3, 1e1])

    f = plt.gcf()
    if legend:
        f.legend(
            handles=[l0, l1, l2, l3],  # The line objects
            labels=[
                "Proton cyclotron",
                "Mirror Mode",
                "Parallel Firehose",
                "Oblique Firehose",
            ],
            loc="upper center",
            borderaxespad=0.1,
            ncol=4,
            frameon=False,
            handlelength=1.0,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.calc_fs import calc_fs
from pyrfu.pyrf.ts_scalar import ts_scalar
from pyrfu.pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def fft_bandpass(inp: DataArray, f_min: float, f_max: float) -> DataArray:
    r"""Perform simple bandpass using FFT - returns fields between with
    ``f_min`` < f < ``f_max``.

    Parameters
    ----------
    inp : DataArray
        Time series to be bandpass filtered.
    f_min : float
        Minimum frequency of filter, f < ``f_min`` are removed.
    f_max : float
        Maximum frequency of filter, f > ``f_max`` are removed.

    Returns
    -------
    DataArray
        Time series of the bandpass filtered data.

    Raises
    ------
    TypeError
        * If input is not a xarray.DataArray.
        * If f_min is not a float.
        * If f_max is not a float.
    ValueError
        If f_min is larger than f_max.

    Notes
    -----
    Can be some spurious effects near boundary. Can take longer interval then
    use pyrfu.pyrf.time_clip to remove.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint = ["2017-07-23T16:54:24.000", "2017-07-23T17:00:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load Electric Field

    >>> e_xyz = mms.get_data("e_gse_edp_brst_l2", tint, mms_id)

    Bandpass filter

    >>> e_xyz_bp = mms.fft_bandpass(e_xyz, 1e1, 1e2)

    """

    # Make sure input is a DataArray
    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a xarray.DataArray")

    # Check f_min and f_max
    if not isinstance(f_min, float):
        raise TypeError("f_min must be a float")

    if not isinstance(f_max, float):
        raise TypeError("f_max must be a float")

    # Check that f_min < f_max
    if f_min >= f_max:
        raise ValueError("f_min must be smaller than f_max")

    # Get time and data
    inp_time: NDArray[np.datetime64] = inp.time.data
    inp_data: NDArrayFloats = inp.data
    precision = inp_data.dtype

    # Reshape to column vector if input is a scalar
    if inp_data.ndim == 1:
        # If scalar, reshape to column vector
        inp_data = inp_data[:, np.newaxis]
    elif inp_data.ndim > 2:
        raise ValueError("Input must be a scalar or a vector")

    # Make sure number of elements is an even number, if odd remove last
    # element to make an even number
    if len(inp_time) % 2:
        inp_time = inp_time[:-1]
        inp_data = inp_data[:-1, :]

    # Set NaN values to zero so FFT works
    idx_nans = np.isnan(inp_data)
    inp_data[idx_nans] = 0.0

    # Bandpass filter field data
    f_sam = calc_fs(inp)
    f_nyq = f_sam / 2
    frequencies = np.linspace(-f_nyq, f_nyq, len(inp_time))

    # Preallocate output
    out_data_64: NDArray[np.float64] = np.empty_like(inp_data, dtype=np.float64)

    # FFT and remove frequencies
    for i in range(inp_data.shape[1]):
        inp_tmp: NDArray[np.float64] = inp_data[:, i].astype(np.float64)
        inp_fft: NDArray[np.complex128] = np.fft.fft(inp_tmp)
        inp_fft = np.fft.fftshift(inp_fft)

        inp_fft[np.abs(frequencies) < f_min] = 0.0 + 0.0j
        inp_fft[np.abs(frequencies) > f_max] = 0.0 + 0.0j

        inp_fft = np.fft.ifftshift(inp_fft)
        out_tmp = np.fft.ifft(inp_fft)

        out_data_64[:, i] = np.real(out_tmp)

    # Put back original NaNs and back to original shape and precision
    out_data_64[idx_nans] = np.nan
    out_data_64 = np.squeeze(out_data_64)
    out_data: NDArrayFloats = out_data_64.astype(precision)

    # Return data in the same format as input
    if out_data.ndim == 1:
        out: DataArray = ts_scalar(inp_time, out_data, attrs=inp.attrs)
    else:
        out = ts_vec_xyz(inp_time, out_data, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np

# Local imports
from ..pyrf.ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def vdf_elim(vdf, e_int):
    r"""Limits the skymap distribution to the selected energy range.

    Parameters
    ----------
    vdf : xarray.Dataset
        Skymap velocity distribution to clip.
    e_int : list or float
        Energy interval boundaries (list) or energy to slice.

    Returns
    -------
    vdf_e_clipped : xarray.Dataset
        Skymap of the clipped velocity distribution.

    """

    energy = vdf.energy
    unique_etables = np.unique(vdf.energy.data, axis=0)
    # n_etables = 2 for older dta and 1 for newer data
    n_etables = unique_etables.shape[0]

    e_int = list(np.atleast_1d(e_int))
    e_int.sort()

    # energy interval
    if len(e_int) == 2:
        e_levels = []

        for i_etable in range(n_etables):
            # loop over 1 or 2 and saves all the unique indices, i.e. max range
            lower_ = e_int[0] < unique_etables[i_etable, :]
            upper_ = unique_etables[i_etable, :] < e_int[1]
            tmp_elevels = np.where(np.logical_and(lower_, upper_))[0]
            e_levels = np.unique(np.hstack([e_levels, tmp_elevels]))

        e_levels = list(e_levels.astype(np.int64))
        e_min = np.min(energy.data[:, e_levels])
        e_max = np.max(energy.data[:, e_levels])
        logging.info(
            "Effective eint = [%(e_min)5.2f, %(e_max)5.2f]",
            {"e_min": e_min, "e_max": e_max},
        )
        energies = energy.data[:, e_levels]
        data = vdf.data.data[:, e_levels, ...]

    else:
        # pick closest energy level
        e_diff0 = np.abs(energy[0, :] - e_int)
        e_diff1 = np.abs(energy[1, :] - e_int)
        if np.min(e_diff0) < np.min(e_diff1):
            e_diff = e_diff0
        else:
            e_diff = e_diff1

        e_levels = int(np.where(e_diff == np.min(e_diff))[0][0])
        logging.info(
            "Effective energies alternate in time between %(e0)5.2f and %(e1)5.2f",
            {"e0": energy.data[0, e_levels], "e1": energy.data[1, e_levels]},
        )
        energies = energy.data[:, e_levels]
        energies = energies[:, np.newaxis]
        data = vdf.data.data[:, e_levels, ...]
        data = data[:, np.newaxis, ...]

    # Data attributes
    data_attrs = vdf.data.attrs

    # Coordinates attributes
    coords_attrs = {k: vdf[k].attrs for k in ["time", "energy", "phi", "theta"]}

    # Global attributes
    glob_attrs = vdf.attrs

    # Get energies levels
    energy_0 = np.atleast_1d(glob_attrs.get("energy0", unique_etables[0, :])[e_levels])
    energy_1 = np.atleast_1d(glob_attrs.get("energy1", unique_etables[0, :])[e_levels])
    esteptable = glob_attrs.get("esteptable", np.zeros(len(vdf.time)))

    vdf_e_clipped = ts_skymap(
        vdf.time.data,
        data,
        energies,
        vdf.phi.data,
        vdf.theta.data,
        energy0=energy_0,
        energy1=energy_1,
        esteptable=esteptable,
        attrs=data_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    return vdf_e_clipped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os
from typing import Mapping, Optional, Tuple, Union

# 3rd party imports
import requests
from botocore.exceptions import ClientError
from requests.sessions import Session
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from pyrfu.mms.db_init import MMS_CFG_PATH
from pyrfu.mms.get_dist import get_dist
from pyrfu.mms.get_ts import get_ts
from pyrfu.mms.list_files import list_files
from pyrfu.mms.list_files_aws import list_files_aws
from pyrfu.mms.list_files_sdc import _login_lasp, list_files_sdc
from pyrfu.mms.tokenize import tokenize
from pyrfu.pyrf.dist_append import dist_append
from pyrfu.pyrf.ts_append import ts_append
from pyrfu.pyrf.ttns2datetime64 import ttns2datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _var_and_cdf_name(
    var_str: str, mms_id: str
) -> Tuple[Mapping[str, Union[str, int]], str]:
    r"""Tokenize variable string `var_str` and returns the corresponding cdf name.

    Parameters
    ----------
    var_str : str
        Variable string.
    mms_id : str
        Spacecraft identifier.

    Returns
    -------
    var : dict
        Variable dictionary.
    cdf_name : str
        Corresponding cdf name.

    """
    var = tokenize(var_str)
    cdf_name = f"mms{mms_id}_{var['cdf_name']}"
    return var, cdf_name


def _check_times(inp: Union[DataArray, Dataset]) -> Union[DataArray, Dataset]:
    if inp.time.data.dtype == "int64":
        out = inp.assign_coords(time=ttns2datetime64(inp.time.data))
    else:
        out = inp
    return out


def _list_files_sources(
    source: str,
    tint: list[str],
    mms_id: str,
    var: Mapping[str, Union[str, int]],
    data_path: Optional[str] = "",
):
    if source == "local":
        file_names = list_files(tint, mms_id, var, data_path)
        sdc_session, headers = None, {}
    elif source == "sdc":
        file_names = [file.get("url") for file in list_files_sdc(tint, mms_id, var)]
        sdc_session, headers, _ = _login_lasp()
    elif source == "aws":
        file_names = [file.get("s3_obj") for file in list_files_aws(tint, mms_id, var)]
        sdc_session, headers = None, {}
    else:
        raise NotImplementedError(f"Resource {source} is not yet implemented!!")

    return file_names, sdc_session, headers


def _get_file_content_sources(
    source: str,
    file_name: str,
    sdc_session: Optional[Session] = None,
    headers: Optional[dict] = None,
) -> bytes:
    r"""Get file content from different sources.

    Parameters
    ----------
    source : str
        Source of the data.
    file_name : str
        File name.
    sdc_session : Session, Optional
        SDC session. Default is None.
    headers : dict, Optional
        Headers. Default is None.

    Returns
    -------
    file_content : bytes
        File content.

    """
    if source == "local":
        file_path = os.path.normpath(file_name)
        with open(file_path, "rb") as file:
            file_content = file.read()
    elif source == "sdc":
        try:
            response = sdc_session.get(file_name, timeout=None, headers=headers)
            response.raise_for_status()  # Raise an HTTPError for bad responses
            file_content = response.content
        except requests.RequestException:
            logging.error("Error retrieving file from %s", file_name)
    elif source == "aws":
        try:
            response = file_name.get()
            file_content = response["Body"].read()
        except ClientError as err:
            if err.response["Error"]["Code"] == "InternalError":  # Generic error
                logging.error("Error Message: %s", err.response["Error"]["Message"])

                response_meta = err.response.get("ResponseMetadata")
                logging.error("Request ID: %s", response_meta.get("RequestId"))
                logging.error("Http code: %s", response_meta.get("HTTPStatusCode"))
            else:
                raise err
    else:
        raise NotImplementedError(f"Resource {source} is not yet implemented!!")

    return file_content


def get_data(
    var_str: str,
    tint: list,
    mms_id: Union[int, str],
    verbose: bool = True,
    data_path: str = "",
    source: str = "",
) -> Union[DataArray, Dataset]:
    r"""Load a variable `var_str`.

    AFG: b_bcs_afg_srvy_l2pre, b_dmpa_afg_srvy_l2pre, b_gse_afg_srvy_l2pre,
    b_gsm_afg_srvy_l2pre

    DFG: b_bcs_dfg_srvy_l2pre, b_dmpa_dfg_srvy_l2pre, b_gse_dfg_srvy_l2pre,
     b_gsm_dfg_srvy_l2pre

    EDP: e2d_dsl_edp_brst_l2pre, e2d_dsl_edp_brst_ql, e2d_dsl_edp_fast_l2pre,
    e2d_dsl_edp_fast_ql, e_dsl_edp_brst_l2, e_dsl_edp_brst_l2pre, e_dsl_edp_brst_ql,
    e_dsl_edp_fast_l2, e_dsl_edp_fast_l2pre, e_dsl_edp_fast_ql, e_dsl_edp_slow_l2,
    e_dsl_edp_slow_l2pre, e_gse_edp_brst_l2, e_gse_edp_fast_l2, e_gse_edp_slow_l2,
    e_ssc_edp_brst_l2a, e_ssc_edp_fast_l2a, e_ssc_edp_slow_l2a, hmfe_dsl_edp_brst_l2,
    phase_edp_fast_l2a, phase_edp_slow_l2a, sdev12_edp_fast_l2a, sdev12_edp_slow_l2a,
    sdev34_edp_fast_l2a, sdev34_edp_slow_l2a, v_edp_brst_l2, v_edp_fast_l2,
    v_edp_fast_sitl, v_edp_slow_l2, v_edp_slow_sitl

    FGM: b_bcs_fgm_brst_l2, b_bcs_fgm_srvy_l2, b_dmpa_fgm_brst_l2, b_dmpa_fgm_srvy_l2,
    b_gse_fgm_brst_l2, b_gse_fgm_srvy_l2, b_gsm_fgm_brst_l2, b_gsm_fgm_srvy_l2

    FPI: defbgi_fpi_brst_l2, defbgi_fpi_fast_l2, defe_fpi_brst_l2, defe_fpi_fast_l2,
    defe_fpi_fast_ql, defi_fpi_brst_l2, defi_fpi_fast_l2, nbgi_fpi_brst_l2,
    nbgi_fpi_fast_l2, ne_fpi_brst_l2, ne_fpi_fast_l2, ne_fpi_fast_ql, ni_fpi_brst_l2,
    ni_fpi_fast_l2, ni_fpi_fast_ql, partne_fpi_brst_l2, partne_fpi_fast_l2,
    partni_fpi_brst_l2, partni_fpi_fast_l2, partpe_gse_fpi_brst_l2,
    partpe_gse_fpi_fast_l2, partpi_gse_fpi_brst_l2, partpi_gse_fpi_fast_l2,
    partte_dbcs_fpi_brst_l2, partte_dbcs_fpi_fast_l2, partte_gse_fpi_brst_l2,
    partte_gse_fpi_fast_l2, partti_dbcs_fpi_brst_l2, partti_dbcs_fpi_fast_l2,
    partti_gse_fpi_brst_l2, partti_gse_fpi_fast_l2, parttparae_fpi_brst_l2,
    parttparai_fpi_brst_l2, parttparai_fpi_fast_l2, parttperpe_fpi_brst_l2,
    parttperpi_fpi_brst_l2, parttperpi_fpi_fast_l2, partve_dbcs_fpi_brst_l2,
    partve_dbcs_fpi_fast_l2, partve_gse_fpi_brst_l2, partve_gse_fpi_fast_l2,
    partvi_dbcs_fpi_brst_l2, partvi_dbcs_fpi_fast_l2, partvi_gse_fpi_brst_l2,
    partvi_gse_fpi_fast_l2, pbgi_fpi_brst_l2, pbgi_fpi_fast_l2, pde_fpi_brst_l2,
    pde_fpi_fast_l2, pderre_fpi_brst_l2, pderre_fpi_fast_l2, pderri_fpi_brst_l2,
    pderri_fpi_fast_l2, pdi_fpi_brst_l2, pdi_fpi_fast_l2, pe_dbcs_fpi_brst_l2,
    pe_dbcs_fpi_fast_l2, pe_dbcs_fpi_fast_ql, pe_gse_fpi_brst_l2, pe_gse_fpi_fast_l2,
    pe_gse_fpi_fast_ql, pi_dbcs_fpi_brst_l2, pi_dbcs_fpi_fast_l2, pi_dbcs_fpi_fast_ql,
    pi_gse_fpi_brst_l2, pi_gse_fpi_fast_l2, pi_gse_fpi_fast_ql, ste_dbcs_fpi_brst_l2,
    ste_dbcs_fpi_fast_l2, ste_gse_fpi_brst_l2, ste_gse_fpi_fast_l2,
    sti_dbcs_fpi_brst_l2, sti_dbcs_fpi_fast_l2, sti_gse_fpi_brst_l2,
    sti_gse_fpi_fast_l2, te_dbcs_fpi_brst_l2, te_dbcs_fpi_fast_l2,
    te_dbcs_fpi_fast_ql, te_gse_fpi_brst_l2, te_gse_fpi_fast_l2, te_gse_fpi_fast_ql,
    ti_dbcs_fpi_brst_l2, ti_dbcs_fpi_fast_l2, ti_gse_fpi_brst_l2, ti_gse_fpi_fast_l2,
    tparae_fpi_brst_l2, tparai_fpi_brst_l2, tparai_fpi_fast_l2, tperpe_fpi_brst_l2,
    tperpi_fpi_brst_l2, tperpi_fpi_fast_l2, ve_dbcs_fpi_brst_l2, ve_dbcs_fpi_fast_l2,
    ve_dbcs_fpi_fast_ql, ve_gse_fpi_brst_l2, ve_gse_fpi_fast_l2, ve_gse_fpi_fast_ql,
    vi_dbcs_fpi_brst_l2, vi_dbcs_fpi_fast_l2, vi_dbcs_fpi_fast_ql,
    vi_gse_fpi_brst_l2, vi_gse_fpi_fast_l2, vi_gse_fpi_fast_ql

    FSM: b_gse_fsm_brst_l3

    HPCA: azimuth_hpca_brst_l2, azimuth_hpca_srvy_l2, dpfheplus_hpca_brst_l2,
    dpfheplus_hpca_srvy_l2, dpfheplusplus_hpca_brst_l2, dpfheplusplus_hpca_srvy_l2,
    dpfhplus_hpca_brst_l2, dpfhplus_hpca_srvy_l2, dpfoplus_hpca_brst_l2,
    dpfoplus_hpca_srvy_l2, nheplus_hpca_brst_l2, nheplus_hpca_srvy_l2,
    nheplusplus_hpca_brst_l2, nheplusplus_hpca_srvy_l2, nhplus_hpca_brst_l2,
    nhplus_hpca_srvy_l2, noplus_hpca_brst_l2, noplus_hpca_srvy_l2, saz_hpca_brst_l2,
    saz_hpca_srvy_l2, theplus_dbcs_hpca_brst_l2, theplus_dbcs_hpca_srvy_l2,
    theplusplus_dbcs_hpca_brst_l2, theplusplus_dbcs_hpca_srvy_l2,
    thplus_dbcs_hpca_brst_l2, thplus_dbcs_hpca_srvy_l2, toplus_dbcs_hpca_brst_l2,
    toplus_dbcs_hpca_srvy_l2, tsheplus_hpca_brst_l2, tsheplus_hpca_srvy_l2,
    tsheplusplus_hpca_brst_l2, tsheplusplus_hpca_srvy_l2, tshplus_hpca_brst_l2,
    tshplus_hpca_srvy_l2, tsoplus_hpca_brst_l2, tsoplus_hpca_srvy_l2,
    vheplus_dbcs_hpca_brst_l2, vheplus_dbcs_hpca_srvy_l2, vheplus_gsm_hpca_brst_l2,
    vheplus_gsm_hpca_srvy_l2, vheplusplus_dbcs_hpca_brst_l2,
    vheplusplus_dbcs_hpca_srvy_l2, vheplusplus_gsm_hpca_brst_l2,
    vheplusplus_gsm_hpca_srvy_l2, vhplus_dbcs_hpca_brst_l2, vhplus_dbcs_hpca_srvy_l2,
    vhplus_gsm_hpca_brst_l2, vhplus_gsm_hpca_srvy_l2, voplus_dbcs_hpca_brst_l2,
    voplus_dbcs_hpca_srvy_l2, voplus_gsm_hpca_brst_l2, voplus_gsm_hpca_srvy_l2

    MEC: r_gse_mec_srvy_l2, r_gsm_mec_srvy_l2, r_gse_mec_brst_l2, r_gsm_mec_brst_l2,
    v_gse_mec_srvy_l2, v_gsm_mec_srvy_l2, v_gse_mec_brst_l2, v_gsm_mec_brst_l2

    SCM: b_gse_scm_brst_l2

    Parameters
    ----------
    var_str : str
        Key of the target variable (use mms.get_data() to see keys.).
    tint : list of str
        Time interval.
    mms_id : str or int
        Index of the target spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Local path of MMS data. Default uses that provided in `pyrfu/mms/config.json`
    source: {"local", "sdc", "aws"}, Optional
        Ressource to fetch data from. Default uses default in `pyrfu/mms/config.json`

    Returns
    -------
    out : DataArray or Dataset
        Time series of the target variable of measured by the target
        spacecraft over the selected time interval.

    See also
    --------
    pyrfu.mms.get_ts : Read time series.
    pyrfu.mms.get_dist : Read velocity distribution function.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Index of MMS spacecraft

    >>> ic = 1

    Load magnetic field from FGM

    >>> b_xyz = mms.get_data("b_gse_fgm_brst_l2", tint_brst, ic)

    """
    # Convert mms_id to string
    mms_id = str(mms_id)

    var, cdf_name = _var_and_cdf_name(var_str, mms_id)

    # Read the current version of the MMS configuration file
    with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
        config = json.load(fs)

    source = source if source else config.get("default")

    file_names, sdc_session, headers = _list_files_sources(
        source, tint, mms_id, var, data_path
    )

    if not file_names:
        raise FileNotFoundError(f"No files found for {var_str} in {source}")

    if verbose:
        logging.info("Loading %s...", cdf_name)

    out = None

    for file_name in file_names:
        file_content = _get_file_content_sources(
            source, file_name, sdc_session, headers
        )

        if "-dist" in var["dtype"]:
            out = dist_append(out, get_dist(file_content, cdf_name, tint))

        else:
            out = ts_append(out, get_ts(file_content, cdf_name, tint))

    out = _check_times(out)

    if sdc_session:
        sdc_session.close()

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings
from copy import copy

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

energies_ = {
    "electron": np.array(
        [
            33.2,
            51.90,
            70.6,
            89.4,
            107.1,
            125.2,
            146.5,
            171.3,
            200.2,
            234.0,
            273.4,
            319.4,
            373.2,
            436.0,
            509.2,
        ],
    ),
    "ion": np.array(
        [
            57.9,
            76.8,
            95.4,
            114.1,
            133.0,
            153.7,
            177.6,
            205.1,
            236.7,
            273.2,
            315.4,
            363.8,
            419.7,
            484.2,
            558.6,
        ],
    ),
}


def feeps_omni(inp_dataset):
    r"""Calculates the omni-directional FEEPS spectrogram.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset with all active telescopes data.

    Returns
    -------
    flux_omni : xarray.DataArray
        Omni-directional FEEPS spectrogram.

    Notes
    -----
    The dataset can be raw data, but it is better to remove bad datas,
    sunlight contamination and split before.

    See Also
    --------
    pyrfu.mms.get_feeps_alleyes, pyrfu.mms.feeps_remove_bad_data,
    pyrfu.mms.feeps_split_integral_ch, pyrfu.mms.feeps_remove_sun

    """
    d_type, specie = [inp_dataset.attrs["dtype"] for _ in range(2)]
    mms_id = inp_dataset.attrs["mmsId"]
    energies = copy(energies_[d_type])

    # set unique energy bins per spacecraft; from DLT on 31 Jan 2017
    e_corr = {
        "electron": [14.0, -1.0, -3.0, -3.0],
        "ion": [0.0, 0.0, 0.0, 0.0],
    }

    g_fact = {"electron": [1.0, 1.0, 1.0, 1.0], "ion": [0.84, 1.0, 1.0, 1.0]}

    energies += e_corr[specie][mms_id - 1]

    top_sensors = list(filter(lambda x: "top" in x, inp_dataset))
    bot_sensors = list(filter(lambda x: "bot" in x, inp_dataset))

    dalleyes = np.empty(
        (
            len(inp_dataset.time),
            len(energies),
            len(top_sensors) + len(bot_sensors),
        ),
    )
    dalleyes[:] = np.nan

    for idx, sensor in enumerate(top_sensors):
        data = inp_dataset[sensor].data
        dalleyes[:, :, idx] = data

    for idx, sensor in enumerate(bot_sensors):
        data = inp_dataset[sensor].data
        dalleyes[:, :, idx + len(top_sensors)] = data

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        flux_omni = np.nanmean(dalleyes, axis=2)

    flux_omni *= g_fact[specie][mms_id - 1]

    attrs = {
        "dtype": inp_dataset.attrs["dtype"],
        "UNITS": inp_dataset[top_sensors[0]].attrs["UNITS"],
        "species": inp_dataset.attrs["species"],
        "mmsId": inp_dataset.attrs["mmsId"],
        "units_name": inp_dataset.attrs["units_name"],
    }

    flux_omni = xr.DataArray(
        flux_omni,
        coords=[inp_dataset.time.data, energies],
        dims=["time", "energy"],
        attrs=attrs,
    )

    return flux_omni



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

# Local imports
from .feeps_energy_table import feeps_energy_table

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_correct_energies(feeps_alle):
    r"""Modifies the energy table in FEEPS spectra (intensity, count_rate,
    counts) using the function: mms_feeps_energy_table (which is s/c, sensor
    head and sensor ID dependent)

    Parameters
    ----------
    feeps_alle : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the
        Fly's Eye Energetic Particle Spectrometer (FEEPS).

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the
        Fly's Eye Energetic Particle Spectrometer (FEEPS) with corrected
        energy table.

    """

    mms_id = feeps_alle.attrs["mmsId"]

    sensors_eyes = list(filter(lambda x: x[:3] in ["top", "bot"], feeps_alle))

    out_dict = {k: feeps_alle[k] for k in feeps_alle if k not in sensors_eyes}

    for sensors_eye in sensors_eyes:
        sensor, eye = sensors_eye.split("-")

        new_energy = feeps_energy_table(mms_id, sensor, int(eye))

        dim = feeps_alle[sensors_eye].dims[1]
        out_dict[sensors_eye] = feeps_alle[sensors_eye].assign_coords(
            {dim: new_energy},
        )
        out_dict[sensors_eye] = out_dict[sensors_eye].rename(
            {dim: f"energy-{sensors_eye}"},
        )

    out = xr.Dataset(out_dict)
    out.attrs = feeps_alle.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
from scipy import constants
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def _mass_ratio(inp):
    if inp.attrs["species"].lower() in ["ions", "ion", "protons", "proton"]:
        mass_ratio = 1
    elif inp.attrs["species"].lower() in ["alphas", "alpha", "helium"]:
        mass_ratio = 4
    elif inp.attrs["species"].lower() in ["electrons", "e"]:
        mass_ratio = constants.electron_mass / constants.proton_mass
    else:
        raise ValueError("Invalid specie")

    return mass_ratio


def _convert(inp, units, mass_ratio):
    fact = 1e6 * 0.53707 * mass_ratio**2

    if units.lower() in ["kev/(cm^2 s sr kev)", "ev/(cm^2 s sr ev)", "1/(cm^2 s sr)"]:
        tmp_data = inp / 1e18 * fact
    else:
        raise ValueError("Invalid unit")

    return tmp_data


def def2psd(inp: Union[DataArray, Dataset]) -> Union[DataArray, Dataset]:
    r"""Compute phase space density from differential energy flux.

    The phase-space density is given by:

    .. math:

        f(E) = m^2 \frac{DEF}{E^2} * 0.53707,

    where :math:`m` is the particle mass in atomic mass unit, :math:`DEF` is
    the differential energy flux in 1/(cm sr s) and :math:`E` is the energy
    in eV.

    Parameters
    ----------
    inp : xarray.Dataset or xarray.DataArray
        Time series of the differential energy flux in [(cm^{2} s sr)^{-1}].

    Returns
    -------
    psd : xarray.Dataset or xarray.DataArray
        Time series of the phase space density in [s^{3} m^{-6}]

    Raises
    ------
    TypeError
        If inp is not a xarray.Dataset or xarray.DataArray.

    """
    if isinstance(inp, Dataset):
        tmp_data = _convert(inp.data.data, inp.data.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy[:, :, None, None], (1, 1, *tmp_data.shape[2:]))
        tmp_data /= energy_mat**2
        out = inp.copy()
        out.data.data = np.squeeze(tmp_data)
        out.data.attrs["UNITS"] = "s^3/m^6"
    elif isinstance(inp, DataArray):
        tmp_data = _convert(inp.data, inp.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy, (tmp_data.shape[0], 1))
        tmp_data /= energy_mat**2
        out = inp.copy()
        out.data = np.squeeze(tmp_data)
        out.attrs["UNITS"] = "s^3/m^6"
    else:
        raise TypeError("Invalid input type")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def calc_sqrtq(p_xyz: DataArray) -> DataArray:
    r"""Computes agyrotropy coefficient as in [1]_

    .. math::

        Q = \frac{P_{12}^2 + P_{13}^2 + P_{23}^2}
        {P_\perp^2 + 2 P_\perp P_\parallel}


    Parameters
    ----------
    p_xyz : DataArray
        Time series of the pressure tensor

    Returns
    -------
    DataArray
        Time series of the agyrotropy coefficient of the specie

    Raises
    ------
    TypeError
        If input is not a xarray.DataArray.
    ValueError
        If input is not a time series of a tensor (n_time, 3, 3).

    References
    ----------
    .. [1]  Swisdak, M. (2016), Quantifying gyrotropy in magnetic
            reconnection, Geophys. Res.Lett., 43, 43–49, doi:
            https://doi.org/10.1002/2015GL066980.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> ic = 1

    Load magnetic field and electron pressure tensor

    >>> b_xyz = mms.get_data("b_gse_fgm_srvy_l2", tint, 1)
    >>> p_xyz_e = mms.get_data("pe_gse_fpi_fast_l2", tint, 1)

    Rotate electron pressure tensor to field aligned coordinates

    >>> p_fac_e_pp = mms.rotate_tensor(p_xyz_e, "fac", b_xyz, "pp")

    Compute agyrotropy coefficient

    >>> sqrt_q_e = pyrf.calc_sqrtq(p_fac_e_pp)

    """

    # Check input type
    if not isinstance(p_xyz, xr.DataArray):
        raise TypeError("p_xyz must be a xarray.DataArray")

    # Check import shape
    if not p_xyz.data.ndim == 3 or p_xyz.shape[1] != 3 or p_xyz.shape[2] != 3:
        raise ValueError("p_xyz must be a time series of a tensor")

    # Parallel and perpendicular components
    p_para: np.ndarray = p_xyz.data[:, 0, 0]
    p_perp: np.ndarray = (p_xyz.data[:, 1, 1] + p_xyz.data[:, 2, 2]) / 2

    # Off-diagonal terms
    p_12: np.ndarray = p_xyz.data[:, 0, 1]
    p_13: np.ndarray = p_xyz.data[:, 0, 2]
    p_23: np.ndarray = p_xyz.data[:, 1, 2]

    sqrt_q: np.ndarray = np.sqrt(p_12**2 + p_13**2 + p_23**2)
    sqrt_q /= np.sqrt(p_perp**2 + 2 * p_perp * p_para)
    sqrt_q_ts: DataArray = ts_scalar(p_xyz.time.data, sqrt_q)

    return sqrt_q_ts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants, optimize

# Local imports
from ..pyrf.resample import resample
from ..pyrf.trace import trace
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_tensor_xyz import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _f_one_pop(x, *args):
    i_e, i_aspoc, sc_pot_r = args
    return np.nansum(
        np.abs(
            i_e.data + i_aspoc.data - (x[0] * np.exp(-sc_pot_r.data / x[1])),
        ),
    )


def _f_two_pop(x, *args):
    i_e, i_aspoc, sc_pot_r = args
    return np.nansum(
        np.abs(
            i_e.data
            + i_aspoc.data
            - (x[0] * np.exp(-sc_pot_r.data / x[1]))
            + (x[2] * np.exp(-sc_pot_r.data / x[3])),
        ),
    )


def scpot2ne(sc_pot, n_e, t_e, i_aspoc: xr.DataArray = None):
    r"""Compute number density from spacecraft potential. Function uses number
    density and electron temperature to compute electron thermal current. A
    best fit of the photoelectron current to the thermal current is used to
    determine the photoelectron currents and temperatures. These are then used
    to construct a new number density from the spacecraft potential.

    Parameters
    ----------
    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.
    n_e : xarray.DataArray
        Time series of the electron number density.
    t_e : xarray.DataArray
        Time series of the electron temperature. Function accepts scalar
        temperature or tensor (scalar temperature is used in either case).
    i_aspoc : xarray.DataArray, Optional
        Time series of the ASPOC current in :math:`\\mu` A.


    Returns
    -------
    n_esc : xarray.DataArray
        Time series of the number density estimated from SCpot, at the same
        resolution as ``sc_pot``.
    i_ph0 : float
        Value of the photoelectron currents ( :math:`\\mu` A) of the first
        population.
    t_ph0 : float
        Value of the temperature (eV) of the first population.
    i_ph1 : float
        Value of the photoelectron currents ( :math:`\\mu` A) of the second
        population.
    t_ph1 : float
        Value of the temperature (eV) of the second population.

    Notes
    -----
    Usual assumptions are made for thermal and photoelectron current, vis.,
    planar geometry for photoelectrons and spherical geometry for thermal
    electrons.
    Currently the calculation neglects the ion thermal current, secondary
    electrons, and other sources of current.
    ASPOC on does not work very well.

    """

    if i_aspoc is not None:
        i_aspoc = resample(i_aspoc, n_e)
    else:
        i_aspoc = ts_scalar(n_e.time.data, np.zeros(len(n_e)))

    # Check format of electron temperature (tensor)
    assert t_e.ndim == 3 and t_e.shape[1:] == (3, 3), "te must be timeseries of tensor"
    t_e = trace(ts_tensor_xyz(t_e.time.data, t_e.data)) / 3

    # Define constants
    m_e = constants.electron_mass
    q_e = constants.elementary_charge
    s_surf = 34

    v_eth = np.sqrt(2 * q_e * t_e.data / m_e)

    sc_pot_r = resample(sc_pot, n_e)

    # Thermal current in muA
    i_e = n_e.data * v_eth * (1 + sc_pot_r.data / t_e.data)
    i_e *= 1e12 * q_e * s_surf / (2 * np.sqrt(np.pi))

    # First a simple fit of Iph to Ie using 1 photoelectron population
    opt_p1 = optimize.fmin(
        _f_one_pop,
        x0=[500.0, 3.0],
        args=(i_e, i_aspoc, sc_pot_r),
        maxfun=5000,
    )

    # Fit of Iph to Ie for two photoelectron populations
    opt_p2 = optimize.fmin(
        _f_two_pop,
        x0=[opt_p1[0], opt_p1[1], 10.0, 10.0],
        args=(i_e, i_aspoc, sc_pot_r),
        maxfun=5000,
    )
    i_ph0, t_ph0, i_ph1, t_ph1 = opt_p2

    v_eth = np.sqrt(2 * q_e * resample(t_e, sc_pot).data / m_e)
    n_esc = i_ph0 * np.exp(-sc_pot.data / t_ph0) + i_ph1 * np.exp(
        -sc_pot.data / t_ph1,
    )
    n_esc /= s_surf * q_e * v_eth * (1 + sc_pot.data / resample(t_e, sc_pot).data)
    n_esc *= 2 * np.sqrt(np.pi) * 1e-12
    n_esc = ts_scalar(sc_pot.time.data, n_esc)

    return n_esc, i_ph0, t_ph0, i_ph1, t_ph1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from pyrfu.pyrf.avg_4sc import avg_4sc
from pyrfu.pyrf.c_4_grad import c_4_grad
from pyrfu.pyrf.c_4_k import c_4_k
from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.trace import trace
from pyrfu.pyrf.ts_scalar import ts_scalar
from pyrfu.pyrf.ts_tensor_xyz import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["pid_4sc", "pid_4sc_err"]


def pid_4sc(r_mms, v_mms, p_mms):
    r"""Compute Pi-D term using definition of [1]_ as :

    .. math::

        Pi-D = - \Pi_{ij}D_{ij}

    with :math:`\Pi_{ij}` the deviatoric part of the pressure
    tensor :

    .. math::

        \Pi_{ij} = P_{ij} - p\delta_{ij}

        p = \frac{1}{3}P_{ii}


    and :math:`D_{ij}` the deviatoric part of the strain tensor :

    .. math::

        D_{ij} = \frac{1}{2}\left ( \partial_i u_j + \partial_j u_i \right )
        - \frac{1}{3}\theta\delta_{ij}

        \theta = \nabla . u


    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the position of the 4 spacecraft.
    v_mms : list of xarray.DataArray
        Time series of the bulk velocities of the 4 spacecraft.
    p_mms : list of xarray.DataArray
        Time series of the pressure tensor of the 4 spacecraft.

    Returns
    -------
    pid : xarray.DataArray
        Time series of the Pi-D.

    References
    ----------
    .. [1]  Yang, Y., Matthaeus, W. H., Parashar, T. N., Wu, P., Wan, M.,
            Shi, Y., et al. (2017). Energy transfer channels and turbulence
            cascade in Vlasov-Maxwell turbulence. Physical Review E, 95,
            061201. doi : https://doi.org/10.1103/PhysRevE.95.061201

    """

    # Compute pressure tensor and background magnetic field at the center of
    # mass of the tetrahedron
    p_xyz = avg_4sc(p_mms)

    # Compute divergence and gradient tensor of the bulk velocity. Yang's
    # notation
    theta = c_4_grad(r_mms, v_mms, "div")
    grad_u = c_4_grad(r_mms, v_mms, "grad")

    # Define identity tensor
    identity_3d = np.zeros((len(grad_u), 3, 3))
    np.einsum("jii->ji", identity_3d)[:] = 1

    # strain tensor
    eps_xyz = (grad_u.data + np.transpose(grad_u.data, [0, 2, 1])) / 2

    # Deviatoric part of the strain tensor
    d_xyz = eps_xyz - theta.data[:, np.newaxis, np.newaxis] * identity_3d / 3
    d_xyz = ts_tensor_xyz(v_mms[0].time.data, d_xyz)

    # Isotropic scalar pressure
    press = trace(p_xyz) / 3

    # Deviatoric part of the pressure tensor
    pi_xyz = p_xyz.data - press.data[:, np.newaxis, np.newaxis] * identity_3d
    pi_xyz = ts_tensor_xyz(v_mms[0].time.data, pi_xyz)

    # Compute p\theta
    ptheta = press.data * theta.data
    ptheta = ts_scalar(v_mms[0].time.data, ptheta)

    # Compute Pi-D
    pid = -np.sum(np.sum(pi_xyz.data * d_xyz.data, axis=1), axis=1)
    pid = ts_scalar(v_mms[0].time.data, pid)

    return d_xyz, pi_xyz, ptheta, pid


def pid_4sc_err(r_mms, v_mms, p_mms, errv_mms, errp_mms):
    r"""Compute error propagation for Pi-D term using definition of [1]_ as :

    .. math::
        Pi-D = - \Pi_{ij}D_{ij}
    with :math:`\Pi_{ij}` the deviatoric part of the pressure
    tensor :
    .. math::
        \Pi_{ij} = P_{ij} - p\delta_{ij}

        p = \frac{1}{3}P_{ii}
    and :math:`D_{ij}` the deviatoric part of the strain tensor :
    .. math::
        D_{ij} = \frac{1}{2}\left ( \partial_i u_j + \partial_j u_i \right )
        - \frac{1}{3}\theta\delta_{ij}

        \theta = \nabla . u

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the position of the 4 spacecraft.
    v_mms : list of xarray.DataArray
        Time series of the bulk velocities of the 4 spacecraft.
    p_mms : list of xarray.DataArray
        Time series of the pressure tensor of the 4 spacecraft.
    errv_mms : list of xarray.DataArray
        Time series of the error on the bulk velocities of the 4 spacecraft.
    errp_mms : list of xarray.DataArray
        Time series of the error on the pressure tensor of the 4 spacecraft.

    Returns
    -------
    sigma_pid : xarray.DataArray
        Time series of the error on Pi-D.

    References
    ----------
    .. [1]  Roberts, O. W., Vörös, Z., Torkar, K., Stawarz, J.,
            Bandyopadhyay, R., Gershman, D. J., et al. (2023).
            Estimation of the error in the calculation of the
            pressure-strain term: Application in the terrestrial
            magnetosphere. Journal of Geophysical Research: Space Physics,
            128, e2023JA031565. https://doi.org/10.1029/2023JA031565

    """

    # Compute Pi-D and its components
    d_xyz, pi_xyz, _, pid = pid_4sc(r_mms, v_mms, p_mms)

    # Compute error propagation for the gradient of the velocity field (see c_4_k)
    errv_mms = [resample(errv_i, errv_mms[0]) for errv_i in errv_mms]
    errp_mms = [resample(errp_i, errv_mms[0]) for errp_i in errp_mms]
    r_mms = [resample(r, errv_mms[0]) for r in r_mms]
    p_mms = [resample(p_i, errv_mms[0]) for p_i in p_mms]

    # Compute reciprocal vectors in barycentric coordinates (see c_4_k)
    k_list = c_4_k(r_mms)

    sigma_grad_v_xyz = np.zeros((k_list[0].shape[0], 3, 3))

    for i in range(3):
        for j in range(3):
            for alpha in range(4):
                sigma_grad_v_xyz[:, i, j] += (
                    k_list[alpha].data[:, i] ** 2 * errv_mms[alpha].data[:, j] ** 2
                )

    sigma_grad_v_xyz = np.sqrt(sigma_grad_v_xyz)

    # Compute error propagation for the symmetric part of the velocity gradient tensor
    sigma_d_xyz = np.zeros_like(sigma_grad_v_xyz)

    # Diagonal components (Eq. 15 in Roberts et al., 2023)
    sigma_d_xyz[:, 0, 0] = (
        (4 / 9) * sigma_grad_v_xyz[:, 0, 0] ** 2
        + (1 / 9) * sigma_grad_v_xyz[:, 1, 1] ** 2
        + (1 / 9) * sigma_grad_v_xyz[:, 2, 2] ** 2
    )
    sigma_d_xyz[:, 1, 1] = (
        (1 / 9) * sigma_grad_v_xyz[:, 0, 0] ** 2
        + (4 / 9) * sigma_grad_v_xyz[:, 1, 1] ** 2
        + (1 / 9) * sigma_grad_v_xyz[:, 2, 2] ** 2
    )
    sigma_d_xyz[:, 2, 2] = (
        (1 / 9) * sigma_grad_v_xyz[:, 0, 0] ** 2
        + (1 / 9) * sigma_grad_v_xyz[:, 1, 1] ** 2
        + (4 / 9) * sigma_grad_v_xyz[:, 2, 2] ** 2
    )

    # Off-diagonal components (Eq. 14 in Roberts et al., 2023)
    sigma_d_xyz[:, 0, 1] = (1 / 4) * sigma_grad_v_xyz[:, 0, 1] ** 2 + (
        1 / 4
    ) * sigma_grad_v_xyz[:, 1, 0] ** 2
    sigma_d_xyz[:, 0, 2] = (1 / 4) * sigma_grad_v_xyz[:, 0, 2] ** 2 + (
        1 / 4
    ) * sigma_grad_v_xyz[:, 2, 0] ** 2
    sigma_d_xyz[:, 1, 2] = (1 / 4) * sigma_grad_v_xyz[:, 1, 2] ** 2 + (
        1 / 4
    ) * sigma_grad_v_xyz[:, 2, 1] ** 2
    sigma_d_xyz[:, 1, 0] = sigma_d_xyz[:, 0, 1]  # Symmetry
    sigma_d_xyz[:, 2, 0] = sigma_d_xyz[:, 0, 2]  # Symmetry
    sigma_d_xyz[:, 2, 1] = sigma_d_xyz[:, 1, 2]  # Symmetry
    sigma_d_xyz = np.sqrt(sigma_d_xyz)

    # Error in the center of mass pressure tensor
    sigma_p_xyz = (1 / 4) * np.sqrt(
        errp_mms[0].data ** 2
        + errp_mms[1].data ** 2
        + errp_mms[2].data ** 2
        + errp_mms[3].data ** 2
    )

    # Error in the traceless pressure tensor
    sigma_pi_xyz = np.zeros_like(sigma_p_xyz)
    sigma_pi_xyz[:, 0, 0] = (
        (4 / 9) * sigma_p_xyz[:, 0, 0] ** 2
        + (1 / 9) * sigma_p_xyz[:, 1, 1] ** 2
        + (1 / 9) * sigma_p_xyz[:, 2, 2] ** 2
    )
    sigma_pi_xyz[:, 1, 1] = (
        (1 / 9) * sigma_p_xyz[:, 0, 0] ** 2
        + (4 / 9) * sigma_p_xyz[:, 1, 1] ** 2
        + (1 / 9) * sigma_p_xyz[:, 2, 2] ** 2
    )
    sigma_pi_xyz[:, 2, 2] = (
        (1 / 9) * sigma_p_xyz[:, 0, 0] ** 2
        + (1 / 9) * sigma_p_xyz[:, 1, 1] ** 2
        + (4 / 9) * sigma_p_xyz[:, 2, 2] ** 2
    )
    sigma_pi_xyz[:, 0, 1] = sigma_p_xyz[:, 0, 1] ** 2
    sigma_pi_xyz[:, 0, 2] = sigma_p_xyz[:, 0, 2] ** 2
    sigma_pi_xyz[:, 1, 2] = sigma_p_xyz[:, 1, 2] ** 2
    sigma_pi_xyz[:, 1, 0] = sigma_pi_xyz[:, 0, 1]  # Symmetry
    sigma_pi_xyz[:, 2, 0] = sigma_pi_xyz[:, 0, 2]  # Symmetry
    sigma_pi_xyz[:, 2, 1] = sigma_pi_xyz[:, 1, 2]  # Symmetry
    sigma_pi_xyz = np.sqrt(sigma_pi_xyz)

    sigma_pid_xyz_rel = np.zeros_like(pid.data)
    sigma_pid_xyz_rel = np.sqrt(
        (sigma_d_xyz / d_xyz.data) ** 2 + (sigma_pi_xyz / pi_xyz.data) ** 2
    )
    sigma_pid_xyz = np.abs(pi_xyz.data * d_xyz.data) * sigma_pid_xyz_rel

    sigma_pid = np.sqrt(
        sigma_pid_xyz[:, 0, 0] ** 2
        + sigma_pid_xyz[:, 1, 1] ** 2
        + sigma_pid_xyz[:, 2, 2] ** 2
        + 4 * sigma_pid_xyz[:, 0, 1] ** 2
        + 4 * sigma_pid_xyz[:, 0, 2] ** 2
        + 4 * sigma_pid_xyz[:, 1, 2] ** 2
    )
    sigma_pid = ts_scalar(v_mms[0].time.data, sigma_pid)

    return sigma_pid



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Dict, Union

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2025"
__license__ = "MIT"
__version__ = "2.4.14"
__status__ = "Prototype"

# Coefficients for electron instabilities
COEFFS_E: Dict[float, Dict[str, tuple]] = {
    0.01: {
        "firehose": (-1.23, 0.88),
        "whistler": (0.36, 0.55),
    },
    0.1: {
        "firehose": (-1.32, 0.61),
        "whistler": (1.0, 0.49),
    },
}

# Coefficients for ion instabilities
COEFFS_I: Dict[float, Dict[str, tuple]] = {
    0.01: {
        "proton cyclotron": (0.649, 0.400, 0.0),
        "mirror mode": (1.040, 0.633, -0.012),
        "parallel firehose": (-0.647, 0.583, 0.713),
        "oblique firehose": (-1.447, 1.000, -0.148),
    },
    0.001: {
        "proton cyclotron": (0.437, 0.428, -0.003),
        "mirror mode": (0.801, 0.763, -0.063),
        "parallel firehose": (-0.497, 0.566, 0.543),
        "oblique firehose": (-1.390, 1.005, -0.111),
    },
    0.0001: {
        "proton cyclotron": (0.367, 0.364, 0.011),
        "mirror mode": (0.702, 0.674, -0.009),
        "parallel firehose": (-0.408, 0.529, 0.410),
        "oblique firehose": (-1.454, 1.023, -0.178),
    },
}


def _thresh_e(
    beta_para: Union[float, np.ndarray], s: float, alpha: float
) -> Union[float, np.ndarray]:
    return 1 + s * beta_para**-alpha


def _thresh_i(
    beta_para: Union[float, np.ndarray], a: float, b: float, beta0: float
) -> Union[float, np.ndarray]:
    return 1 + a / (beta_para - beta0) ** b


def anisotropy_thresholds(
    beta_para: Union[float, np.ndarray], specie: str = "i", gamma: float = 0.01
) -> Dict[str, Union[float, np.ndarray]]:
    r"""Compute the thresholds for temperature anisotropy instabilities based on
    plasma species and growth rate.

    Parameters
    ----------
    beta_para : float or array_like
        Parallel beta.
    specie : str, optional
        Plasma species, "i" for ions or "e" for electrons. Default is "i".
    gamma : float, optional
        Growth rate of the instability. Must match a key in the corresponding
        coefficient dictionary. Default is 0.01.

    Returns
    -------
    dict
        Dictionary of thresholds with instability names as keys.

    Raises
    ------
    ValueError
        If specie is not "i" or "e", or if gamma is not supported.

    """

    if specie == "i":
        if gamma not in COEFFS_I:
            gammas = list(COEFFS_I.keys())
            raise ValueError(
                f"Unsupported gamma value {gamma} for ions. Available: {gammas}"
            )
        coeffs = COEFFS_I[gamma]
        out = {name: _thresh_i(beta_para, *params) for name, params in coeffs.items()}

    elif specie == "e":
        if gamma not in COEFFS_E:
            gammas = list(COEFFS_E.keys())
            raise ValueError(
                f"Unsupported gamma value {gamma} for electrons. Available: {gammas}"
            )
        coeffs = COEFFS_E[gamma]
        out = {name: _thresh_e(beta_para, *params) for name, params in coeffs.items()}

    else:
        raise ValueError(f"Unknown specie '{specie}'. Expected 'i' or 'e'.")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .eis_ang_ang import eis_ang_ang

# Local imports
from .eis_combine_proton_spec import eis_combine_proton_spec
from .eis_skymap import eis_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_combine_proton_skymap(
    phxtof_allt,
    extof_allt,
    en_chan: list = None,
    to_psd: bool = True,
):
    r"""Combines ExTOF and PHxTOF proton energy spectra and generate proton
    skymap distribution.

    Parameters
    ----------
    phxtof_allt : xarray.Dataset
        Dataset containing the PHxTOF energy spectrum of the 6 telescopes.
    extof_allt : xarray.Dataset
        Dataset containing the ExTOF energy spectrum of the 6 telescopes.
    en_chan : array_like, Optional
        Energy channels to use. Default use all energy channels.
    to_psd : bool, Optional
        Flag to convert differential particle flux to phase space density.

    Returns
    -------
    eis_skymap : xarray.Dataset
        EIS skymap distribution

    """

    # Combine Pulse-Height x Time Of Flight and Energy x Time Of Flight spectra
    eis_allt = eis_combine_proton_spec(phxtof_allt, extof_allt)

    # Compute EIS angle-angle (azimuth-elevation) distribution.
    eis_ang_ = eis_ang_ang(eis_allt, en_chan=en_chan)

    # Convert to skymap format
    eis_skymap_ = eis_skymap(eis_ang_, to_psd=to_psd)

    return eis_skymap_



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging
from typing import Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from pyrfu.mms.dsl2gse import _transformation_matrix
from pyrfu.pyrf.calc_fs import calc_fs
from pyrfu.pyrf.cotrans import cotrans
from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.ts_tensor_xyz import ts_tensor_xyz
from pyrfu.pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def rotate_tensor(
    inp: DataArray,
    rot_flag: str,
    vec: Union[DataArray, NDArrayFloats, Dataset],
    perp: Optional[str] = "pp",
    verbose: Optional[bool] = False,
) -> DataArray:
    r"""Rotates pressure or temperature tensor into another coordinate system.

    Parameters
    ----------
    inp : DataArray
        Time series of the tensor to rotate. The tensor must be given in the
        form of a 3x3 matrix, where the first index is the time index and the
        last two indices are the tensor indices.

    rot_flag : str
        Flag of the target coordinates system :
            * "fac" : 	Field-aligned coordinates, requires background
                        magnetic field Bback, optional flag "pp"
                        :math:`\\mathbf{P}_{\\perp 1} = \\mathbf{P}_{\\perp2}`
                        or "qq" :math:`\\mathbf{P}_{\\perp 1}` and
                        :math:`\\mathbf{P}_{\\perp 2}` are most unequal, sets
                        P_{23} to zero.

            * "rot" :	Arbitrary coordinate system, requires new x-direction
                        xnew, new y and z directions ynew, znew (if not
                        included y and z directions are orthogonal to xnew
                        and closest to the original y and z directions)

            * "gse" :   GSE coordinates, requires MMS spacecraft attitude DEFATT.

            * "gsm" :   GSM coordinates, requires MMS spacecraft attitude DEFATT.

    vec : DataArray or numpy.ndarray or xarray.Dataset
        Vector or coordinates system to rotate the tensor. If vec is time series of a
        vector tensor is rotated in field aligned coordinates. If vec is a
        numpy.ndarray rotates to a time independent coordinates system.

    perp : str, Optional
        Flag for perpendicular components of the tensor. Default is pp.
            * "pp" : perpendicular diagonal components are equal
            * "qq" : perpendicular diagonal components are most unequal

    verbose : bool, Optional
        Set to True to print additional information.

    Returns
    -------
    DataArray
        Time series of the pressure or temperature tensor in field-aligned,
        user-defined, or GSE coordinates.
        For "fac" Pe = [Ppar P12 P13; P12 Pperp1 P23; P13 P23 Pperp2].
        For "rot" and "gse" Pe = [Pxx Pxy Pxz; Pxy Pyy Pyz; Pxz Pyz Pzz]

    Raises
    ------
    TypeError
        * If inp is not a xarray.DataArray.
        * If vec is not a xarray.DataArray, numpy.ndarray or xarray.Dataset.
        * If rot_flag is not a string.
        * If perp is not a string.
    NotImplementedError
        * If rot_flag is not 'fac', 'rot', 'gse', or 'gsm'.
        * If rot_flag is 'rot' and vec is time dependent.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and ion temperature tensor
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> # Compute ion temperature in field aligned coordinates
    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")

    """
    # Check input
    if not isinstance(inp, xr.DataArray):
        raise TypeError("inp must be a xarray.DataArray")

    if not isinstance(vec, (xr.DataArray, np.ndarray, xr.Dataset)):
        raise TypeError(
            "vec must be a xarray.DataArray, numpy.ndarray or xarray.Dataset"
        )

    if not isinstance(rot_flag, str):
        raise TypeError("rot_flag must be a string")

    if not isinstance(perp, str):
        raise TypeError("perp must be a string")

    # Get input data type
    precision = inp.data.dtype

    # Check input and load pressure/temperature terms
    inp_time: NDArray[np.datetime64] = inp.time.data
    inp_data: NDArrayFloats = inp.data
    n_t: int = len(inp_time)

    # Initialize rotation matrix
    rot_mat: NDArray[np.float64] = np.zeros((n_t, 3, 3), dtype=np.float64)

    if rot_flag.lower() == "fac":
        # If field-aligned coordinates vec must be the background magnetic field,
        # i.e., vector time series DataArray
        if not isinstance(vec, xr.DataArray):
            raise TypeError("vec must be a xarray.DataArray when rot_flag is 'fac'")

        # Resample magnetic field to the same time as the tensor
        b_back: DataArray = resample(vec, inp, f_s=calc_fs(inp))

        # Unit vector of the background magnetic field gives the parallel direction
        b_vec: NDArray[np.float64] = b_back.data.astype(np.float64)
        b_vec /= np.linalg.norm(b_vec, axis=1, keepdims=True)

        # Construct the rotation matrix
        # Parallel direction
        r_x: NDArray[np.float64] = b_vec
        # Perp1 arbitrarily chosen along x
        r_y: NDArray[np.float64] = np.array([1.0, 0.0, 0.0], dtype=np.float64)
        # Perp2 with correction
        r_z: NDArray[np.float64] = np.cross(r_x, r_y)
        r_z /= np.linalg.norm(r_z, axis=1, keepdims=True)
        r_y = np.cross(r_z, r_x)  # Corrected perp1 direction
        r_y /= np.linalg.norm(r_y, axis=1, keepdims=True)

        rot_mat[:, 0, :], rot_mat[:, 1, :], rot_mat[:, 2, :] = [r_x, r_y, r_z]

    elif rot_flag.lower() == "rot":
        # If arbitrary coordinates vec must be the new coordinate system, i.e.,
        # vector or a matrix. Currently only implemented for time independent
        # coordinate system.
        if not isinstance(vec, np.ndarray):
            raise TypeError("vec must be a numpy.ndarray when rot_flag is 'rot'")

        if vec.ndim == 1 and vec.shape[0] == 3:
            # First direction along the input vector
            r_x = vec.astype(np.float64)
            r_x /= np.linalg.norm(r_x, keepdims=True)
            # Second direction arbitrarily chosen along y
            r_y = np.array([0.0, 1.0, 0.0], dtype=np.float64)
            # Third direction orthogonal to x and y directions
            r_z = np.cross(r_x, r_y)
            r_z /= np.linalg.norm(r_z, keepdims=True)
            # Corrected y direction
            r_y = np.cross(r_z, r_x)
            r_y /= np.linalg.norm(r_y, keepdims=True)

        elif vec.ndim == 2 and vec.shape[0] == 3 and vec.shape[1] == 3:
            vec = vec.astype(np.float64)
            # Make sure the rotation matrix is normalized and right-handed.
            r_x = vec[:, 0] / np.linalg.norm(vec[:, 0], keepdims=True)
            r_y = vec[:, 1] / np.linalg.norm(vec[:, 1], keepdims=True)

            r_z = np.cross(r_x, r_y) / np.linalg.norm(np.cross(r_x, r_y))
            r_y = np.cross(r_z, r_x) / np.linalg.norm(np.cross(r_z, r_x))

            # Check if the coordinate system has been changed and issue a warning
            if (
                abs(np.rad2deg(np.arccos(np.dot(r_y, vec[:, 0])))) > 1.0
                or abs(np.rad2deg(np.arccos(np.dot(r_z, vec[:, 2])))) > 1.0
            ):
                logging.warning(
                    "The new coordinate system has been changed to be right handed "
                    "orthogonal.",
                )

        else:
            raise NotImplementedError("Time dependent rotation matrix not implemented.")

        rot_mat[:, 0, :] = np.ones((n_t, 1), dtype=np.float64) * r_x
        rot_mat[:, 1, :] = np.ones((n_t, 1), dtype=np.float64) * r_y
        rot_mat[:, 2, :] = np.ones((n_t, 1), dtype=np.float64) * r_z

    elif rot_flag.lower() in ["gse", "gsm"]:
        # If GSE or GSM coordinates vec must be the spacecraft attitude, i.e.,
        # Dataset with z_ra and z_dec variables
        if not isinstance(vec, xr.Dataset):
            raise TypeError(
                "vec must be a xarray.Dataset when rot_flag is 'gse' or 'gsm'"
            )

        z_ra: NDArray[np.float64] = np.deg2rad(vec.z_ra.data.astype(np.float64))
        z_dec: NDArray[np.float64] = np.deg2rad(vec.z_dec.data.astype(np.float64))

        # Compute the spin axis direction in Geocentric equatorial inertial (GEI)
        # coordinates
        x: NDArray[np.float64] = np.cos(np.deg2rad(z_dec)) * np.cos(np.deg2rad(z_ra))
        y: NDArray[np.float64] = np.cos(np.deg2rad(z_dec)) * np.sin(np.deg2rad(z_ra))
        z: NDArray[np.float64] = np.sin(np.deg2rad(z_dec))
        sax_gei: DataArray = ts_vec_xyz(
            vec.time.data, np.transpose(np.vstack([x, y, z]))
        )

        # Compute the spin axis direction in the new coordinate system
        sax_rot: DataArray = cotrans(sax_gei, f"gei>{rot_flag.lower()}")
        spin_ax_rot: DataArray = resample(sax_rot, inp, f_s=calc_fs(inp))

        # Compute the rotation matrix
        rot_mat = _transformation_matrix(spin_ax_rot.data, 1)

    else:
        raise NotImplementedError(f"{rot_flag} method is not implemented.")

    p_tensor_p: NDArrayFloats = np.zeros((n_t, 3, 3), dtype=precision)

    for i in range(n_t):
        rot_temp: NDArrayFloats = np.squeeze(rot_mat[i, :, :].astype(precision))

        p_tensor_p[i, :, :] = np.matmul(
            np.matmul(rot_temp, np.squeeze(inp_data[i, :, :])),
            np.transpose(rot_temp),
        )

    if perp.lower() == "" or rot_flag.lower() in ["gse", "gsm"]:
        # maybe also add "rot" here??
        logging.info("No additional rotation applied.")
    elif perp.lower() == "pp":
        if verbose:
            logging.info(
                "Applying additional rotation to make the perpendicular components "
                "most equal"
            )
        thetas: NDArrayFloats = 0.5 * np.arctan(
            (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]) / (2 * p_tensor_p[:, 1, 2]),
        )
        thetas[np.isnan(thetas)] = 0.0

        for i, theta in enumerate(thetas):
            rot_temp = np.array(
                [
                    [1, 0, 0],
                    [0, np.cos(theta), np.sin(theta)],
                    [0, -np.sin(theta), np.cos(theta)],
                ],
            )

            p_tensor_p[i, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[i, :, :])),
                np.transpose(rot_temp),
            )

    elif perp.lower() == "qq":

        if verbose:
            logging.info(
                "Applying additional rotation to make the perpendicular components "
                "most unequal"
            )

        thetas = 0.5 * np.arctan(
            (2 * p_tensor_p[:, 1, 2]) / (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]),
        )

        for i, theta in enumerate(thetas):
            rot_temp = np.array(
                [
                    [1, 0, 0],
                    [0, np.cos(theta), -np.sin(theta)],
                    [0, np.sin(theta), np.cos(theta)],
                ],
            )

            p_tensor_p[i, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[i, :, :])),
                np.transpose(rot_temp),
            )
    else:
        raise NotImplementedError(f"{perp} method is not implemented.")

    # Construct output
    p_new: DataArray = ts_tensor_xyz(inp_time, p_tensor_p, attrs=inp.attrs)

    return p_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

# Local imports
from ..pyrf.plasma_calc import plasma_calc

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def whistler_b2e(b2, freq, theta_k, b_mag, n_e):
    r"""Computes electric field power as a function of frequency for whistler
    waves using magnetic field power and cold plasma theory.

    Parameters
    ----------
    b2 : xarray.DataArray
        Time series of the power of whistler magnetic field in nT^2 Hz^{-1}.
    freq : ndarray
        frequencies in Hz corresponding B2.
    theta_k : float
        wave-normal angle of whistler waves in radians.
    b_mag : xarray.DataArray
        Time series of the magnitude of the magnetic field in nT.
    n_e : xarray.DataArray
        Time series of the electron number density in cm^{-3}.

    Returns
    -------
    e2 : xarray.DataArray
        Time series of the electric field power.

    Examples
    --------
    >>> from pyrfu import mms
    >>> e_power = mms.whistler_b2e(b_power, freq, theta_k, b_mag, n_e)

    """

    # Calculate plasma parameters
    pparam = plasma_calc(b_mag, n_e, n_e, n_e, n_e)
    fpe, fce = [pparam.Fpe, pparam.Fce]

    # Check input
    if len(b2) != len(freq):
        raise IndexError("B2 and freq lengths do not agree!")

    # Calculate cold plasma parameters
    rr = 1 - fpe**2 / (freq * (freq - fce))
    ll = 1 - fpe**2 / (freq * (freq + fce))
    pp = 1 - fpe**2 / freq**2
    dd = 0.5 * (rr - ll)
    ss = 0.5 * (rr + ll)

    n2 = rr * ll * np.sin(theta_k) ** 2
    n2 += pp * ss * (1 + np.cos(theta_k) ** 2)
    n2 -= np.sqrt(
        (rr * ll - pp * ss) ** 2 * np.sin(theta_k) ** 4
        + 4 * (pp**2) * (dd**2) * np.cos(theta_k) ** 2,
    )
    n2 /= 2 * (ss * np.sin(theta_k) ** 2 + pp * np.cos(theta_k) ** 2)

    e_temp1 = (pp - n2 * np.sin(theta_k) ** 2) ** 2.0 * ((dd / (ss - n2)) ** 2 + 1) + (
        n2 * np.cos(theta_k) * np.sin(theta_k)
    ) ** 2
    e_temp2 = (dd / (ss - n2)) ** 2 * (
        pp - n2 * np.sin(theta_k) ** 2
    ) ** 2 + pp**2 * np.cos(theta_k) ** 2

    e2 = (constants.speed_of_light**2 / n2) * e_temp1 / e_temp2 * b2.data
    e2 *= 1e-12

    return e2
